/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/window.m: Implementation of SILWindow, a subclass of
 * NSWindow with fullscreen and OpenGL support.
 */

#define IN_SYSDEP

#import "src/base.h"
#import "src/input.h"
#import "src/math.h"
#import "src/memory.h"
#import "src/sysdep.h"
#import "src/sysdep/darwin/time.h"
#import "src/sysdep/opengl/opengl.h"
#import "src/sysdep/macosx/input.h"
#import "src/sysdep/macosx/util.h"
#import "src/sysdep/macosx/view.h"
#import "src/sysdep/macosx/window.h"

#import "src/sysdep/macosx/osx-headers.h"
#import <AppKit/NSCursor.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSMenu.h>
#import <AppKit/NSOpenGL.h>
#import <AppKit/NSScreen.h>
#import <AppKit/NSTextInputContext.h>
#import <AppKit/NSTrackingArea.h>
#import <AppKit/NSWindow.h>
#import <AppKit/NSWindowRestoration.h>
#import <OpenGL/OpenGL.h>

/*************************************************************************/
/************************ SILWindow implementation ***********************/
/*************************************************************************/

@interface SILWindow: NSWindow <NSWindowDelegate> {
  @public
    BOOL setup_ok;  // Did -[setup] succeed?
}

/**
 * setup:  Perform post-initialization setup of the window.  On success,
 * the setup_ok field will be set to YES.
 *
 * This function must be called on the main thread.
 *
 * [Parameters]
 *     format: Pixel format for the window.
 */
- (void)setup:(NSOpenGLPixelFormat *)format;

/**
 * setFullscreen:  Set whether to display the window as a fullscreen window.
 *
 * [Parameters]
 *     fullscreen: True to display the window in fullscreen mode, false to
 *         display it as a normal window.
 *     screen: Screen on which the window is to be displayed.
 */
- (void)setFullscreen:(BOOL)fullscreen screen:(NSScreen *)screen;
/* Wrappers for calling via -[performSelector]. */
- (void) setFullscreen:(NSScreen *)screen;
- (void) setWindowed:(NSScreen *)screen;

/**
 * setShowPointer:  Set whether to display the mouse pointer.
 */
- (void)setShowPointer:(BOOL)show;

/**
 * getInputContext:  Return the window's input context.
 */
- (NSTextInputContext *)getInputContext;

/**
 * hasFocus:  Return whether the window currently has input focus.
 */
- (BOOL)hasFocus;

/**
 * isMoving:  Return whether the window is currently moving.
 */
- (BOOL)isMoving;

/**
 * resize:  Resize the window to the given size.
 *
 * [Parameters]
 *     size: New size.
 * [Return value]
 *     True on success, false on error.
 */
- (BOOL)resize:(NSSize)size;

@end

/*-----------------------------------------------------------------------*/

@implementation SILWindow {
    NSUInteger base_style;   // Style flags passed to initializer.
    BOOL is_fullscreen;      // Is the window currently in fullscreen mode?
    BOOL has_focus;          // Does the window currently have focus?
    BOOL is_moving;          // Is the window currently being moved?
    BOOL show_pointer;       // Should we show the mouse pointer?
    BOOL is_pointer_inside;  // Is the pointer currently inside the window?
    NSCursor *empty_cursor;  // Empty image used to hide the mouse pointer.
    NSString *title;         // Window title string.
    NSTextInputContext *ic;  // Input context for key->text translation.
}

/*-----------------------------------------------------------------------*/

- (id)initWithContentRect:(NSRect)contentRect
                styleMask:(NSUInteger)windowStyle
                  backing:(NSBackingStoreType)bufferingType
                    defer:(BOOL)deferCreation
{
    self = [super initWithContentRect:contentRect styleMask:windowStyle
                  backing:bufferingType defer:deferCreation];

    self.delegate = self;

    setup_ok = NO;
    base_style = windowStyle;
    is_fullscreen = NO;
    has_focus = NO;
    is_moving = NO;
    show_pointer = YES;
    is_pointer_inside = NO;
    title = NULL;

    /* We don't worry about failure here because (1) it's not fatal (the
     * worst that happens is the cursor doesn't disappear, or text input
     * doesn't work and we leak an input client instance), and (2) there's
     * no good way to test it anyway. */
    static const uint8_t empty_pixel[4] = {0,0,0,0};
    NSImage *empty_image = macosx_create_image(empty_pixel, 1, 1);
    empty_cursor = [[NSCursor alloc] initWithImage:empty_image
                                     hotSpot:(NSPoint){0,0}];
    [empty_image release];

    id<NSTextInputClient> input_client = macosx_create_input_client();
    ic = [[NSTextInputContext alloc] initWithClient:input_client];

    return self;
}

/*-----------------------------------------------------------------------*/

- (void)close
{
    if (is_fullscreen) {
        [NSMenu setMenuBarVisible:YES];
    }
    [super close];
}

/*-----------------------------------------------------------------------*/

- (void)dealloc
{
    [title release];
    [empty_cursor release];
    [(NSObject<NSTextInputClient> *)ic.client release];
    [ic release];

    [super dealloc];
}

/*-----------------------------------------------------------------------*/

/* OS X seems to make unbalanced calls to retain and release/autorelease
 * on NSWindow objects in a multithreaded environment, so we null out
 * Objective-C memory management to prevent the object from being destroyed
 * until we explicitly destroy it. */

- (id)retain {return self;}
- (oneway void)release {}
- (id)autorelease {return self;}

/* Called by SILWindow_destroy(). */
- (oneway void)destroy {[super release];}

/*-----------------------------------------------------------------------*/

- (void)setup:(NSOpenGLPixelFormat *)format
{
    /* This is required in order to get NSMouseMoved events. */
    self.acceptsMouseMovedEvents = YES;

    /* Window restoration makes no sense for the type of programs with
     * which SIL is likely to be used.  Disable it so users don't get a
     * "do you want to restore this app's windows?" if the program crashes. */
    [self setRestorable:NO];

    CGRect content_rect = SILWindow_content_frame(self);
    content_rect.origin.x = 0;
    content_rect.origin.y = 0;
    SILView *view = [[SILView alloc] initWithFrame:content_rect
                                     pixelFormat:format];
    [format release];  // Retained by the view.
    if (UNLIKELY(!view)) {
        DLOG("Failed to create view");
        return;
    }
    [self setContentView:view];

    NSTrackingArea *tracking_area =
        [[NSTrackingArea alloc] initWithRect:[view frame]
                                options:NSTrackingCursorUpdate
                                        | NSTrackingMouseEnteredAndExited
                                        | NSTrackingActiveInKeyWindow
                                owner:self userInfo:nil];
    [view addTrackingArea:tracking_area];

    [self makeKeyAndOrderFront:self];
    [self makeMainWindow];

    setup_ok = YES;
}

/*-----------------------------------------------------------------------*/

- (void)setTitle:(NSString *)newTitle
{
    if (newTitle != title) {
        [title release];
        title = [newTitle retain];
    }
    if (!is_fullscreen) {
        [super setTitle:title];
    }
}

/*-----------------------------------------------------------------------*/

/* Windows need to respond to these selectors to become fullscreen. */

- (BOOL)canBecomeKeyWindow
{
    return YES;
}

- (BOOL)canBecomeMainWindow
{
    return YES;
}

/*-----------------------------------------------------------------------*/

/* Handle key window notifications to track focus (we could use -[isKeyWindow],
 * but Apple mysteriously deprecated that in OS X 10.10) and update mouse
 * cursor grab state. */

- (void)becomeKeyWindow
{
    [super becomeKeyWindow];
    has_focus = YES;
    macosx_update_mouse_grab(-1);
    /* We apparently also need to refresh the cursor.  The argument here is
     * unused, but the OS X 10.11 SDK complains if we pass nil. */
    [self cursorUpdate:(void *)1];
}

- (void)resignKeyWindow
{
    macosx_update_mouse_grab(0);
    [super resignKeyWindow];
    has_focus = NO;
}

/*-----------------------------------------------------------------------*/

/* NSTrackingArea callbacks.  We show or hide the pointer depending on the
 * current show state, and track whether the pointer is currently inside
 * the window. */

- (void)cursorUpdate:(NSEvent *) UNUSED event
{
    /* When creating a window in fullscreen mode, the tracking regions
     * don't seem to work properly, so we unconditionally hide the pointer
     * for fullscreen windows. */
    if (show_pointer && !is_fullscreen) {
        [[NSCursor arrowCursor] set];
    } else {
        [empty_cursor set];
    }
}

- (void)mouseEntered:(NSEvent *) UNUSED event
{
    is_pointer_inside = YES;
}

- (void)mouseExited:(NSEvent *) UNUSED event
{
    is_pointer_inside = NO;
}

/* Helper for -[setShowPointer:] that only calls [NSCursor set] if the
 * pointer is currently inside the window. */
- (void)updateCursorIfInside
{
    if (is_pointer_inside) {
        /* The argument here is unused, but the OS X 10.11 SDK complains if
         * we pass nil. */
        [self cursorUpdate:(void *)1];
    }
}

/*-----------------------------------------------------------------------*/

/* Generic event processing routine.  We use this to detect the end of a
 * window move operation (see the init routine). */

- (void)sendEvent:(NSEvent *)event
{
    [super sendEvent:event];

    if (is_moving && [event type] == NSLeftMouseUp) {
        is_moving = NO;
    }
}

/*-----------------------------------------------------------------------*/

/* Mouse and keyboard input handlers.  Note that even if we didn't process
 * keyboard input, we'd have to define the keyboard input handlers as
 * no-ops to prevent the system from beeping whenever a key was pressed. */

- (void)mouseDown:        (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)otherMouseDown:   (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)rightMouseDown:   (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)mouseUp:          (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)otherMouseUp:     (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)rightMouseUp:     (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)mouseDragged:     (NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)otherMouseDragged:(NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)rightMouseDragged:(NSEvent *)event {macosx_handle_mouse_event(event);}
- (void)mouseMoved:       (NSEvent *)event {macosx_handle_mouse_event(event);}

- (void)scrollWheel:      (NSEvent *)event {macosx_handle_scroll_event(event);}

- (void)keyDown:          (NSEvent *)event {macosx_handle_key_event(event);}
- (void)keyUp:            (NSEvent *)event {macosx_handle_key_event(event);}
- (void)flagsChanged:     (NSEvent *)event {macosx_handle_key_event(event);}

/*-----------------------------------------------------------------------*/

/* Window delegate callbacks. */

- (BOOL)windowShouldClose:(id) UNUSED sender
{
    macosx_quit_requested = 1;
    return NO;
}

- (void)windowWillMove:(NSNotification *) UNUSED notification
{
    is_moving = YES;
}

/* The windowDidMove notification is sadly useless for us because it only
 * tells us when the window has just moved, not when the user has finished
 * the move operation.  Instead, we do like SDL2 and treat a mouse-up event
 * while moving as the end of the move operation. */

/*-----------------------------------------------------------------------*/

- (void)setFullscreen:(BOOL)fullscreen screen:(NSScreen *)screen
{
    if (fullscreen == is_fullscreen) {
        return;  // No change.
    }

    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    /* OS X 10.7+ has a "fullscreen mode" in which the fullscreen window
     * is placed in a separate space, but that involves a ridiculously
     * long transition which has also been reported to give some people
     * motion sickness, so we just change the window parameters (with a
     * fade applied in graphics.m). */

    if (fullscreen) {
        [NSMenu setMenuBarVisible:NO];
        [self setStyleMask:NSBorderlessWindowMask];
        [self setFrameOrigin:[screen frame].origin];
        [self setHidesOnDeactivate:YES];
        is_fullscreen = YES;
    } else {
        const float width = roundf([self frame].size.width);
        const float height = roundf([self frame].size.height);
        const CGRect cocoa_bounds = screen.visibleFrame;
        const int x = iroundf(cocoa_bounds.origin.x
                              + (cocoa_bounds.size.width - width) / 2);
        const int y = iroundf(cocoa_bounds.origin.y
                              + (cocoa_bounds.size.height - height) / 2);
        [NSMenu setMenuBarVisible:YES];
        [self setStyleMask:base_style];
        [self setFrameOrigin:(NSPoint){x,y}];
        if (title) {
            [super setTitle:title];
        }
        [self setHidesOnDeactivate:NO];
        is_fullscreen = NO;
    }

    [pool release];
}

/* Wrappers for calling via -[performSelector]. */
- (void) setFullscreen:(NSScreen *)screen {
    [self setFullscreen:YES screen:screen];
}
- (void) setWindowed:(NSScreen *)screen {
    [self setFullscreen:NO screen:screen];
}

/*-----------------------------------------------------------------------*/

- (void)setShowPointer:(BOOL)show
{
    if (show != show_pointer) {
        show_pointer = show;
        /* This is probably safe from any thread, but we run it on the
         * main thread to avoid racing with enter/exit events. */
        [self performSelectorOnMainThread:@selector(updateCursorIfInside)
              withObject:nil waitUntilDone:NO];
    }
}

/*-----------------------------------------------------------------------*/

- (NSTextInputContext *)getInputContext
{
    return ic;
}

/*-----------------------------------------------------------------------*/

- (BOOL)hasFocus
{
    return has_focus;
}

/*-----------------------------------------------------------------------*/

- (BOOL)isMoving
{
    return is_moving;
}

/*-----------------------------------------------------------------------*/

- (BOOL)resize:(NSSize)size
{
    self.contentMinSize = size;
    self.contentMaxSize = size;
    [self setContentSize:size];

    NSRect content_frame = [self contentRectForFrameRect:[self frame]];
    return content_frame.size.width == size.width
        && content_frame.size.height == size.height;
}

/*-----------------------------------------------------------------------*/

@end

/*************************************************************************/
/*********************** Internal utility routines ***********************/
/*************************************************************************/

void *SILWindow_create(int x, int y, int width, int height, int screen,
                       int fullscreen, CGLPixelFormatObj pixel_format)
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    NSRect content_rect;
    content_rect.origin.x = x;
    content_rect.origin.y = y;
    content_rect.size.width = width;
    content_rect.size.height = height;
    const int style_mask = NSTitledWindowMask
                         | NSClosableWindowMask
                         | NSMiniaturizableWindowMask;
    NSScreen *screen_obj = [[NSScreen screens] objectAtIndex:screen];
    SILWindow *window = [[SILWindow alloc] initWithContentRect:content_rect
                                           styleMask:style_mask
                                           backing:NSBackingStoreBuffered
                                           defer:NO
                                           screen:screen_obj];
    if (UNLIKELY(!window)) {
        DLOG("Failed to create window");
        goto error_return;
    }

    NSOpenGLPixelFormat *format =
        [[NSOpenGLPixelFormat alloc] initWithCGLPixelFormatObj:pixel_format];
    if (UNLIKELY(!format)) {
        goto error_destroy_window;
    }

    [window performSelectorOnMainThread:@selector(setup:) withObject:format
            waitUntilDone:YES];
    if (UNLIKELY(!window->setup_ok)) {
        goto error_destroy_window;
    }

    /* Create an OpenGL context for the window. */
    ASSERT(![NSOpenGLContext currentContext]);
    if (UNLIKELY(![[window contentView] createGLContext:TRUE])) {
        DLOG("Failed to create OpenGL context");
        goto error_destroy_window;
    }

    if (fullscreen) {
        [window performSelectorOnMainThread:@selector(setFullscreen:)
                withObject:screen_obj waitUntilDone:YES];
    }

    [pool release];
    return window;

  error_destroy_window:
    [window release];
  error_return:
    [pool release];
    return NULL;
}

/*-----------------------------------------------------------------------*/

void SILWindow_destroy(void *window_)
{
    SILWindow *window = (SILWindow *)window_;

    if (!window) {
        return;
    }

    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    [[window contentView] destroyGLContext];
    [window performSelectorOnMainThread:@selector(close) withObject:nil
            waitUntilDone:YES];
    [window destroy];

    [pool release];
}

/*-----------------------------------------------------------------------*/

void *SILWindow_input_context(void *window_)
{
    PRECOND(window_ != NULL, return NULL);
    SILWindow *window = (SILWindow *)window_;

    return [window getInputContext];
}

/*-----------------------------------------------------------------------*/

CGRect SILWindow_frame(void *window_)
{
    PRECOND(window_ != NULL, return (CGRect){{0,0},{0,0}});
    SILWindow *window = (SILWindow *)window_;

    /* [window frame] gives us coordinates with the origin in the lower
     * left corner, but we want to return CoreGraphics-style coordinates
     * with the origin in the upper left corner. */
    return macosx_window_frame_to_CGRect([window frame]);
}

/*-----------------------------------------------------------------------*/

CGRect SILWindow_content_frame(void *window_)
{
    PRECOND(window_ != NULL, return (CGRect){{0,0},{0,0}});
    SILWindow *window = (SILWindow *)window_;

    return macosx_window_frame_to_CGRect(
        [window contentRectForFrameRect:[window frame]]);
}

/*-----------------------------------------------------------------------*/

int SILWindow_is_moving(void *window_)
{
    PRECOND(window_ != NULL, return 0);
    SILWindow *window = (SILWindow *)window_;

    return [window isMoving];
}

/*-----------------------------------------------------------------------*/

int SILWindow_has_focus(void *window_)
{
    PRECOND(window_ != NULL, return 0);
    SILWindow *window = (SILWindow *)window_;

    return [window hasFocus];
}

/*-----------------------------------------------------------------------*/

void SILWindow_set_title(void *window_, const char *title)
{
    PRECOND(window_ != NULL, return);
    PRECOND(title != NULL, return);
    SILWindow *window = (SILWindow *)window_;

    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    [window performSelectorOnMainThread:@selector(setTitle:)
            withObject:[NSString stringWithUTF8String:title]
            waitUntilDone:YES];

    [pool release];
}

/*-----------------------------------------------------------------------*/

char *SILWindow_get_title(void *window_)
{
    PRECOND(window_ != NULL, return NULL);
    SILWindow *window = (SILWindow *)window_;

    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    NSString *title_str = [window title];
    char *title = NULL;
    if (title_str) {
        const char *s = [title_str UTF8String];
        if (UNLIKELY(!s)) {
            DLOG("Failed to extract title from NSString");
        } else {
            title = mem_strdup(s, MEM_ALLOC_TEMP);
            if (UNLIKELY(!title)) {
                DLOG("No memory for copy of title string");
            }
        }
    }

    [pool release];
    return title;
}

/*-----------------------------------------------------------------------*/

void SILWindow_set_fullscreen(void *window_, int fullscreen, int screen)
{
    PRECOND(window_ != NULL, return);
    SILWindow *window = (SILWindow *)window_;

    NSScreen *screen_obj = [[NSScreen screens] objectAtIndex:screen];
    if (!screen_obj) {
        DLOG("Invalid screen: %d", screen);
        screen_obj = [[NSScreen screens] objectAtIndex:0];
    }

    if (fullscreen) {
        [window performSelectorOnMainThread:@selector(setFullscreen:)
                withObject:screen_obj waitUntilDone:YES];
    } else {
        [window performSelectorOnMainThread:@selector(setWindowed:)
                withObject:screen_obj waitUntilDone:YES];
    }
}

/*-----------------------------------------------------------------------*/

void SILWindow_set_show_pointer(void *window_, int show)
{
    PRECOND(window_ != NULL, return);
    SILWindow *window = (SILWindow *)window_;

    [window setShowPointer:(show != 0)];
}

/*-----------------------------------------------------------------------*/

int SILWindow_resize(void *window_, int width, int height)
{
    PRECOND(window_ != NULL, return 0);
    SILWindow *window = (SILWindow *)window_;

    return [window resize:(NSSize){width, height}];
}

/*-----------------------------------------------------------------------*/

int SILWindow_create_gl_shader_compilation_context(void *window_)
{
    PRECOND(window_ != NULL, return 0);
    SILWindow *window = (SILWindow *)window_;

    ASSERT(![NSOpenGLContext currentContext], return 0);
    return [[window contentView] createGLContext:FALSE];
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
