/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/graphics/framebuffer.c: Tests for framebuffer functionality.
 */

#include "src/base.h"
#include "src/framebuffer.h"
#include "src/graphics.h"
#include "src/graphics/internal.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*************************************************************************/
/*************************** Common test data ****************************/
/*************************************************************************/

/* Vertex format for test data. */
typedef struct BasicVertex {float x,y,z;} BasicVertex;
static const uint32_t basic_vertex_format[] = {
    GRAPHICS_VERTEX_FORMAT(POSITION_3F, offsetof(BasicVertex,x)),
    0
};

/* White (RGB=255,255,255) quad over the horizontal center (|x|<0.5) of
 * the render target. */
static const BasicVertex white_quad_vertices[] = {
    {-0.5,-1,0}, {-0.5,1,0}, {0.5,1,0}, {0.5,-1,0}
};

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

static int wrap_test_graphics_framebuffer(void);
static int do_test_graphics_framebuffer(void);
int test_graphics_framebuffer(void)
{
    return run_tests_in_window(wrap_test_graphics_framebuffer);
}

static int wrap_test_graphics_framebuffer(void)
{
    if (!framebuffer_supported()) {
        SKIP("Framebuffers not supported.");
    }
    return do_test_graphics_framebuffer();
}

DEFINE_GENERIC_TEST_RUNNER(do_test_graphics_framebuffer)

TEST_INIT(init)
{
    graphics_start_frame();
    framebuffer_bind(0);
    graphics_set_viewport(0, 0, TESTW, TESTH);
    graphics_clear(0, 0, 0);
    return 1;
}

TEST_CLEANUP(cleanup)
{
    graphics_finish_frame();
    graphics_flush_resources();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create)
{
    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(128, 96));

    CHECK_INTEQUAL(framebuffer_width(framebuffer), 128);
    CHECK_INTEQUAL(framebuffer_height(framebuffer), 96);

    framebuffer_destroy(framebuffer);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_invalid)
{
    CHECK_FALSE(framebuffer_create(0, 96));
    CHECK_FALSE(framebuffer_create(128, 0));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_destroy_invalid)
{
    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(128, 96));
    framebuffer_destroy(framebuffer);

    framebuffer_destroy(0);  // Defined as a no-op.
    /* The rest of these are invalid calls, but check that they don't crash. */
    framebuffer_destroy(-1);
    framebuffer_destroy(framebuffer);
    framebuffer_destroy(INT_MAX);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_width_height_invalid)
{
    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(128, 96));
    framebuffer_destroy(framebuffer);

    CHECK_INTEQUAL(framebuffer_width(0), 0);
    CHECK_INTEQUAL(framebuffer_width(framebuffer), 0);
    CHECK_INTEQUAL(framebuffer_width(INT_MAX), 0);
    CHECK_INTEQUAL(framebuffer_height(0), 0);
    CHECK_INTEQUAL(framebuffer_height(framebuffer), 0);
    CHECK_INTEQUAL(framebuffer_height(INT_MAX), 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_bind)
{
    uint8_t *pixels;

    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(128, 96));

    /* This should clear the display rather than the framebuffer. */
    graphics_clear(0, 0, 1);

    framebuffer_bind(framebuffer);
    graphics_set_viewport(0, 0, 128, 96);
    /* Neither of these should render to the display. */
    graphics_clear(1, 0, 0);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, white_quad_vertices, basic_vertex_format,
        sizeof(*white_quad_vertices), lenof(white_quad_vertices));

    /* texture_grab[_new]() should read out of the currently bound
     * framebuffer rather than the display buffer. */
    int texture;
    CHECK_TRUE(texture = texture_grab_new(0, 0, 128, 96, 1, 0, 0));
    CHECK_INTEQUAL(texture_width(texture), 128);
    CHECK_INTEQUAL(texture_height(texture), 96);
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int i = 0; i < 128*96*4; i += 4) {
        const int x = (i/4) % 128, y = (i/4) / 128;
        const int p = (x >= 32 && x < 96) ? 255 : 0;
        CHECK_PIXEL(&pixels[i], 255,p,p,255, x, y);
    }
    texture_destroy(texture);

    /* This should break the framebuffer binding. */
    framebuffer_destroy(framebuffer);

    /* This should now read out of the display buffer. */
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,0,255,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_bind_invalid)
{
    uint8_t *pixels;

    int framebuffer, framebuffer2;
    CHECK_TRUE(framebuffer = framebuffer_create(128, 96));
    CHECK_TRUE(framebuffer2 = framebuffer_create(128, 96));
    framebuffer_destroy(framebuffer2);

    framebuffer_bind(framebuffer);
    graphics_set_viewport(0, 0, 128, 96);
    graphics_clear(1, 0, 0);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, white_quad_vertices, basic_vertex_format,
        sizeof(*white_quad_vertices), lenof(white_quad_vertices));

    /* None of these should cause the binding to fall back to the display
     * buffer. */
    framebuffer_bind(-1);
    framebuffer_bind(framebuffer2);
    framebuffer_bind(INT_MAX);

    /* This should still be writing to the framebuffer. */
    graphics_clear(0, 0, 1);

    framebuffer_bind(0);
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,0,0,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    framebuffer_bind(framebuffer);
    int texture;
    CHECK_TRUE(texture = texture_grab_new(0, 0, 128, 96, 1, 0, 0));
    CHECK_INTEQUAL(texture_width(texture), 128);
    CHECK_INTEQUAL(texture_height(texture), 96);
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int i = 0; i < 128*96*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,0,255,255, (i/4) % 128, (i/4) / 128);
    }
    texture_destroy(texture);

    framebuffer_destroy(framebuffer);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_texture)
{
    uint8_t *pixels;

    graphics_clear(0, 0, 1);

    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(64, 64));

    framebuffer_bind(framebuffer);
    graphics_set_viewport(0, 0, 64, 64);
    graphics_clear(1, 0, 0);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, white_quad_vertices, basic_vertex_format,
        sizeof(*white_quad_vertices), lenof(white_quad_vertices));

    framebuffer_bind(0);
    graphics_set_viewport(0, 0, TESTW, TESTH);
    texture_apply(0, framebuffer_get_texture(framebuffer));
    graphics_set_blend(GRAPHICS_BLEND_ADD, 1, 0);

    const float x0 = -1, y0 = -1;
    const float x1 = -1 + 2*(64.0f/TESTW);
    const float y1 = -1 + 2*(64.0f/TESTH);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                   &(Vector2f){0,0}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                   &(Vector2f){0,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                   &(Vector2f){1,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                   &(Vector2f){1,0}, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        if ((x == 16-1 || x == 16 || x == 48-1 || x == 48) && y < 64) {
            continue;  // Skip possibly-antialiased pixels.
        }
        const int r = (x < 64 && y < 64) ? 255 : 0;
        const int g = ((x >= 16 && x < 48) && y < 64) ? 255 : 0;
        const int b = ((x >= 16 && x < 48) || x >= 64 || y >= 64) ? 255 : 0;
        CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
    }
    mem_free(pixels);

    texture_apply(0, 0);

    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                   &(Vector2f){0,0}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                   &(Vector2f){0,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                   &(Vector2f){1,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                   &(Vector2f){1,0}, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        const int r = (x < 64 && y < 64) ? 255 : 0;
        const int g = (x < 64 && y < 64) ? 255 : 0;
        const int b = 255;
        CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
    }
    mem_free(pixels);

    framebuffer_destroy(framebuffer);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_get_texture_invalid)
{
    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(64, 64));
    framebuffer_destroy(framebuffer);

    CHECK_FALSE(framebuffer_get_texture(0));
    CHECK_FALSE(framebuffer_get_texture(framebuffer));
    CHECK_FALSE(framebuffer_get_texture(INT_MAX));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_antialias)
{
    uint8_t *pixels;

    graphics_clear(0, 0, 1);

    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(32, 32));

    framebuffer_bind(framebuffer);
    graphics_set_viewport(0, 0, 32, 32);
    graphics_clear(1, 0.2, 0.2);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, white_quad_vertices, basic_vertex_format,
        sizeof(*white_quad_vertices), lenof(white_quad_vertices));

    framebuffer_bind(0);
    graphics_set_viewport(0, 0, TESTW, TESTH);
    texture_apply(0, framebuffer_get_texture(framebuffer));
    graphics_set_blend(GRAPHICS_BLEND_ADD, 1, 0);

    /* Default (antialiasing on). */
    const float x0 = -1, y0 = -1;
    const float x1 = -1 + 2*(64.0f/TESTW);
    const float y1 = -1 + 2*(64.0f/TESTH);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                   &(Vector2f){0,0}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                   &(Vector2f){0,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                   &(Vector2f){1,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                   &(Vector2f){1,0}, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        if ((x == 16-1 || x == 16 || x == 48-1 || x == 48) && y < 64) {
            if (pixels[i+0] != 255
             || (pixels[i+1] <= 51 || pixels[i+1] >= 255)
             || (pixels[i+2] <= 51 || pixels[i+2] >= 255)
             ||  pixels[i+3] != 255) {
                FAIL("Pixel (%d,%d) was RGBA (%u,%u,%u,%u) but should have"
                     " been between (%u,%u,%u,%u) and (%u,%u,%u,%u)", x, y,
                     pixels[i+0], pixels[i+1], pixels[i+2], pixels[i+3],
                     255, 51, 51, 255, 255, 255, 255, 255);
            }
        } else {
            const int r = (x < 64 && y < 64) ? 255 : 0;
            const int g = (x < 64 && y < 64
                           ? (x > 16 && x < 48-1 ? 255 : 51)
                           : 0);
            const int b = (x < 64 && y < 64
                           ? (x > 16 && x < 48-1 ? 255 : 51)
                           : 255);
            CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
        }
    }
    mem_free(pixels);

    /* On -> off and off -> off transition. */
    for (int try = 0; try < 2; try++) {
        framebuffer_set_antialias(framebuffer, 0);
        CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                       &(Vector2f){0,0}, NULL));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                       &(Vector2f){0,1}, NULL));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                       &(Vector2f){1,1}, NULL));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                       &(Vector2f){1,0}, NULL));
        CHECK_TRUE(graphics_end_and_draw_primitive());
        ASSERT(pixels = grab_display());
        for (int i = 0; i < TESTW*TESTH*4; i += 4) {
            const int x = (i/4) % TESTW;
            const int y = (i/4) / TESTW;
            const int r = (x < 64 && y < 64) ? 255 : 0;
            const int g = ((x >= 16 && x < 48) && y < 64) ? 255 :
                          (x < 64 && y < 64) ? 51 : 0;
            const int b = ((x >= 16 && x < 48) || x >= 64 || y >= 64)
                          ? 255 : 51;
            CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
        }
        mem_free(pixels);
    }

    /* Off -> on and on -> on transition. */
    for (int try = 0; try < 2; try++) {
        framebuffer_set_antialias(framebuffer, 1);
        CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                       &(Vector2f){0,0}, NULL));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                       &(Vector2f){0,1}, NULL));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                       &(Vector2f){1,1}, NULL));
        CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                       &(Vector2f){1,0}, NULL));
        CHECK_TRUE(graphics_end_and_draw_primitive());
        ASSERT(pixels = grab_display());
        for (int i = 0; i < TESTW*TESTH*4; i += 4) {
            const int x = (i/4) % TESTW;
            const int y = (i/4) / TESTW;
            if ((x == 16-1 || x == 16 || x == 48-1 || x == 48) && y < 64) {
                if (pixels[i+0] != 255
                 || (pixels[i+1] <= 51 || pixels[i+1] >= 255)
                 || (pixels[i+2] <= 51 || pixels[i+2] >= 255)
                 ||  pixels[i+3] != 255) {
                    FAIL("Pixel (%d,%d) was RGBA (%u,%u,%u,%u) but should have"
                         " been between (%u,%u,%u,%u) and (%u,%u,%u,%u)", x, y,
                         pixels[i+0], pixels[i+1], pixels[i+2], pixels[i+3],
                         255, 51, 51, 255, 255, 255, 255, 255);
                }
            } else {
                const int r = (x < 64 && y < 64) ? 255 : 0;
                const int g = (x < 64 && y < 64
                               ? (x > 16 && x < 48-1 ? 255 : 51)
                               : 0);
                const int b = (x < 64 && y < 64
                               ? (x > 16 && x < 48-1 ? 255 : 51)
                               : 255);
                CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
            }
        }
        mem_free(pixels);
    }

    framebuffer_destroy(framebuffer);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_antialias_invalid)
{
    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(64, 64));
    framebuffer_destroy(framebuffer);

    framebuffer_set_antialias(0, 0);
    framebuffer_set_antialias(framebuffer, 0);
    framebuffer_set_antialias(INT_MAX, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_discard)
{
    uint8_t *pixels;

    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(TESTW, TESTH));
    framebuffer_bind(framebuffer);
    graphics_set_viewport(0, 0, TESTW, TESTH);
    graphics_clear(1, 0, 0);
    framebuffer_discard_data(framebuffer);
    graphics_clear(0, 1, 0);
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,255,0,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);
    framebuffer_destroy(framebuffer);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_discard_other_framebuffer)
{
    uint8_t *pixels;
    int framebuffer1, framebuffer2;
    CHECK_TRUE(framebuffer1 = framebuffer_create(TESTW, TESTH));
    CHECK_TRUE(framebuffer2 = framebuffer_create(TESTW, TESTH));

    graphics_set_viewport(0, 0, TESTW, TESTH);
    framebuffer_bind(0);
    graphics_clear(1, 0, 0);
    framebuffer_discard_data(framebuffer1);
    graphics_clear(0, 1, 0);  // Should not draw to framebuffer1.
    framebuffer_bind(0);
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,255,0,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    framebuffer_bind(framebuffer1);
    graphics_clear(0, 0, 1);
    framebuffer_discard_data(framebuffer2); // Should not invalidate framebuffer1.
    framebuffer_bind(framebuffer1);
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,0,255,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    framebuffer_bind(framebuffer1);
    graphics_clear(1, 0, 1);
    framebuffer_discard_data(framebuffer2);
    graphics_clear(0, 1, 1);  // Should not draw to framebuffer2.
    framebuffer_bind(framebuffer1);
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,255,255,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    framebuffer_destroy(framebuffer1);
    framebuffer_destroy(framebuffer2);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_discard_invalid)
{
    int framebuffer;
    CHECK_TRUE(framebuffer = framebuffer_create(64, 64));
    framebuffer_destroy(framebuffer);

    framebuffer_discard_data(0);
    framebuffer_discard_data(framebuffer);
    framebuffer_discard_data(INT_MAX);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_memory_failure)
{
    uint8_t *pixels;

    graphics_clear(0, 0, 1);

    /* Prime any arrays that may be extended by the call (e.g., on PSP),
     * so we don't get a bogus memory leak error. */
    framebuffer_destroy(framebuffer_create(64, 64));

    int framebuffer;
    CHECK_MEMORY_FAILURES(framebuffer = framebuffer_create(64, 64));

    CHECK_INTEQUAL(framebuffer_width(framebuffer), 64);
    CHECK_INTEQUAL(framebuffer_height(framebuffer), 64);

    framebuffer_bind(framebuffer);
    graphics_set_viewport(0, 0, 64, 64);
    graphics_clear(1, 0, 0);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, white_quad_vertices, basic_vertex_format,
        sizeof(*white_quad_vertices), lenof(white_quad_vertices));

    framebuffer_bind(0);
    graphics_set_viewport(0, 0, TESTW, TESTH);
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 0,0,255,255, (i/4) % TESTW, (i/4) / TESTW);
    }
    mem_free(pixels);

    framebuffer_set_antialias(framebuffer, 0);
    texture_apply(0, framebuffer_get_texture(framebuffer));
    graphics_set_blend(GRAPHICS_BLEND_ADD, 1, 0);
    const float x0 = -1, y0 = -1;
    const float x1 = -1 + 2*(64.0f/TESTW);
    const float y1 = -1 + 2*(64.0f/TESTH);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                   &(Vector2f){0,0}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                   &(Vector2f){0,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                   &(Vector2f){1,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                   &(Vector2f){1,0}, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        const int r = (x < 64 && y < 64) ? 255 : 0;
        const int g = ((x >= 16 && x < 48) && y < 64) ? 255 : 0;
        const int b = ((x >= 16 && x < 48) || x >= 64 || y >= 64) ? 255 : 0;
        CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
    }
    mem_free(pixels);

    framebuffer_destroy(framebuffer);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_multiple_framebuffers)
{
    int framebuffer1, framebuffer2;
    uint8_t *pixels;

    graphics_clear(0, 0, 1);

    CHECK_TRUE(framebuffer1 = framebuffer_create(64, 64));
    CHECK_INTEQUAL(framebuffer_width(framebuffer1), 64);
    CHECK_INTEQUAL(framebuffer_height(framebuffer1), 64);
    framebuffer_set_antialias(framebuffer1, 0);
    framebuffer_bind(framebuffer1);

    CHECK_TRUE(framebuffer2 = framebuffer_create(32, 32));
    CHECK_INTEQUAL(framebuffer_width(framebuffer2), 32);
    CHECK_INTEQUAL(framebuffer_height(framebuffer2), 32);
    framebuffer_set_antialias(framebuffer2, 0);

    graphics_set_viewport(0, 0, 64, 64);
    graphics_clear(1, 0, 0);
    graphics_draw_vertices(
        GRAPHICS_PRIMITIVE_QUADS, white_quad_vertices, basic_vertex_format,
        sizeof(*white_quad_vertices), lenof(white_quad_vertices));

    framebuffer_bind(framebuffer2);
    graphics_set_viewport(0, 0, 32, 32);
    graphics_clear(0, 1, 0);

    framebuffer_bind(framebuffer1);
    graphics_set_viewport(0, 0, 64, 64);
    texture_apply(0, framebuffer_get_texture(framebuffer2));
    graphics_set_blend(GRAPHICS_BLEND_ADD, 1, 0);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,-1,0},
                                   &(Vector2f){0,0}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,0,0},
                                   &(Vector2f){0,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){0,0,0},
                                   &(Vector2f){1,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){0,-1,0},
                                   &(Vector2f){1,0}, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    framebuffer_discard_data(framebuffer2);

    framebuffer_bind(0);
    graphics_set_viewport(0, 0, TESTW, TESTH);
    texture_apply(0, framebuffer_get_texture(framebuffer1));
    graphics_set_blend(GRAPHICS_BLEND_ADD, 1, 0);
    const float x0 = -1, y0 = -1;
    const float x1 = -1 + 2*(64.0f/TESTW);
    const float y1 = -1 + 2*(64.0f/TESTH);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y0,0},
                                   &(Vector2f){0,0}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x0,y1,0},
                                   &(Vector2f){0,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y1,0},
                                   &(Vector2f){1,1}, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){x1,y0,0},
                                   &(Vector2f){1,0}, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        const int r = (x < 32 && y < 32) ? 0 :
                      (x < 64 && y < 64) ? 255 : 0;
        const int g = (x < 32 && y < 32) ? 255 :
                      ((x >= 16 && x < 48) && y < 64) ? 255 : 0;
        const int b = (x < 32 && y < 32) ? 0 :
                      ((x >= 16 && x < 48) || x >= 64 || y >= 64) ? 255 : 0;
        CHECK_PIXEL(&pixels[i], r,g,b,255, x, y);
    }
    mem_free(pixels);

    framebuffer_destroy(framebuffer1);
    framebuffer_destroy(framebuffer2);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_hole_in_array)
{
    int framebuffer1, framebuffer2, framebuffer3;

    CHECK_TRUE(framebuffer1 = framebuffer_create(64, 64));
    CHECK_TRUE(framebuffer2 = framebuffer_create(32, 32));

    framebuffer_destroy(framebuffer1);

    CHECK_INTEQUAL(framebuffer_create(16, 16), framebuffer1);
    CHECK_TRUE(framebuffer3 = framebuffer_create(8, 8));

    CHECK_INTEQUAL(framebuffer_width(framebuffer1), 16);
    CHECK_INTEQUAL(framebuffer_height(framebuffer1), 16);
    CHECK_INTEQUAL(framebuffer_width(framebuffer2), 32);
    CHECK_INTEQUAL(framebuffer_height(framebuffer2), 32);
    CHECK_INTEQUAL(framebuffer_width(framebuffer3), 8);
    CHECK_INTEQUAL(framebuffer_height(framebuffer3), 8);

    framebuffer_destroy(framebuffer2);
    framebuffer_cleanup();

    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
