/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/graphics/misc.c: Miscellaneous tests for graphics functions.
 */

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_graphics_misc)

TEST_INIT(init)
{
    return graphics_init();
}

TEST_CLEANUP(cleanup)
{
    graphics_cleanup();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_multisample)
{
    if (!(graphics_set_display_attr("multisample", 4))) {
        SKIP("Multisample not supported.");
    }
    if (!open_window(TESTW, TESTH)) {
        SKIP("Unable to set a multisample display mode.");
    }
    graphics_set_viewport(0, 0, TESTW, TESTH);

    graphics_start_frame();

    graphics_clear(0, 0, 0);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,0,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,1,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){1.0f/TESTW,1,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){1.0f/TESTW,0,0}, NULL, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    uint8_t *pixels;
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        if (x == TESTW/2 && y >= TESTH/2) {
            /* Sample points are generally pseudorandom, so we don't know
             * exactly what value we'll get back from the renderer, but it
             * should be a non-black, non-white grey. */
            const int p = pixels[i];
            if (p == 0 || p == 255 || pixels[i+1] != p || pixels[i+2] != p
             || pixels[i+3] != 255) {
                FAIL("Pixel (%d,%d) was RGBA (%u,%u,%u,%u) but should have"
                     " been multisampled grey",
                     x, y, pixels[i+0], pixels[i+1], pixels[i+2], pixels[i+3]);
            }
        } else {
            const int p = (x < TESTW/2 && y >= TESTH/2) ? 255 : 0;
            CHECK_PIXEL(&pixels[i], p,p,p,255, x, y);
        }
    }
    mem_free(pixels);

    graphics_finish_frame();
    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
