/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/linux/graphics.c: Linux-specific graphics-related tests.
 */

/* These tests cannot be built with Clang because the system call wrappers
 * they require make use of the GCC __builtin_apply() set of functions,
 * which are not supported by Clang. */
#ifdef __clang__
#include "src/base.h"
#include "src/test/base.h"
int test_linux_graphics(void) {SKIP("Not available for Clang builds.");}
#else  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/sysdep.h"
#include "src/sysdep/linux/internal.h"
#include "src/sysdep/opengl/dyngl.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/test/sysdep/linux/wrap-x11.h"
#include "src/texture.h"

#include <dlfcn.h>
#define RTLD_DEFAULT  NULL
#define RTLD_NEXT     ((void *)(intptr_t)-1)

#include <time.h>
#include <GL/glx.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/extensions/Xinerama.h>
#include <X11/extensions/Xrandr.h>
#include <X11/extensions/xf86vmode.h>

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * get_alternate_mode:  Return a non-default video mode usable for testing
 * fullscreen behavior at alternate resolutions.
 *
 * If the environment variable SIL_TEST_ALTERNATE_DISPLAY_MODE is set to a
 * string like "1280x720", that resolution is used unconditionally.  This
 * can be used to work around systems that claim to provide video modes
 * which they don't actually support.
 *
 * [Parameters]
 *     width_ret: Pointer to variable to receive the display mode width.
 *     height_ret: Pointer to variable to receive the display mode height.
 * [Return value]
 *     True if an alternate mode was found, false if not.
 */
static int get_alternate_mode(int *width_ret, int *height_ret)
{
    const char *env = getenv("SIL_TEST_ALTERNATE_DISPLAY_MODE");
    if (env && *env) {
        const char *s;
        *width_ret = (int)strtol(env, (char **)&s, 10);
        if (*s == 'x') {
            *height_ret = (int)strtol(s+1, (char **)&s, 10);
            if (!*s) {
                return 1;
            }
        }
        static int warned = 0;
        if (!warned) {
            WARN("Invalid syntax for SIL_TEST_ALTERNATE_DISPLAY_MODE: %s\n",
                 env);
            warned = 1;
        }
    }

    return get_alternate_video_mode(0, width_ret, height_ret);
}

/*-----------------------------------------------------------------------*/

/**
 * check_vidmode:  Return whether the XF86VidMode extension is available.
 */
static int check_vidmode(void)
{
    __typeof__(XF86VidModeQueryExtension) *p_XF86VidModeQueryExtension =
        dlsym(RTLD_NEXT, "XF86VidModeQueryExtension");
    int event_base, error_base;
    return p_XF86VidModeQueryExtension
        && (*p_XF86VidModeQueryExtension)(linux_x11_display(),
                                          &event_base, &error_base);
}

/*-----------------------------------------------------------------------*/

/**
 * check_xinerama:  Return whether the Xinerama extension is available and
 * active.
 */
static int check_xinerama(void)
{
    __typeof__(XineramaQueryExtension) *p_XineramaQueryExtension =
        dlsym(RTLD_NEXT, "XineramaQueryExtension");
    __typeof__(XineramaIsActive) *p_XineramaIsActive =
        dlsym(RTLD_NEXT, "XineramaIsActive");
    int event_base, error_base;
    return p_XineramaQueryExtension
        && p_XineramaIsActive
        && (*p_XineramaQueryExtension)(linux_x11_display(),
                                       &event_base, &error_base)
        && (*p_XineramaIsActive)(linux_x11_display());
}

/*-----------------------------------------------------------------------*/

/**
 * check_xrandr:  Return whether XRandR 1.2+ is available.
 */
static int check_xrandr(void)
{
    __typeof__(XRRQueryExtension) *p_XRRQueryExtension =
        dlsym(RTLD_NEXT, "XRRQueryExtension");
    __typeof__(XRRQueryVersion) *p_XRRQueryVersion =
        dlsym(RTLD_NEXT, "XRRQueryVersion");
    int event_base, error_base, major, minor;
    return p_XRRQueryExtension
        && p_XRRQueryVersion
        && (*p_XRRQueryExtension)(linux_x11_display(), &event_base, &error_base)
        && (*p_XRRQueryVersion)(linux_x11_display(), &major, &minor)
        && (major >= 2 || (major == 1 && minor >= 2));
}

/*-----------------------------------------------------------------------*/

/**
 * xrandr_get_current_resolution:  Return the current screen resolution for
 * the screen containing the current window.  XRandR must be available, and
 * the window must have been opened on SIL device index 0.
 *
 * [Parameters]
 *     width_ret: Pointer to variable to receive the screen width, in pixels.
 *     height_ret: Pointer to variable to receive the screen height, in pixels.
 */
static void xrandr_get_current_resolution(int *width_ret, int *height_ret)
{
    PRECOND(width_ret);
    PRECOND(height_ret);
    PRECOND(linux_x11_window());

    /* We don't have wrappers for these, so we need to look them up. */
    __typeof__(XRRFreeScreenResources) *p_XRRFreeScreenResources =
        dlsym(RTLD_DEFAULT, "XRRFreeScreenResources");
    __typeof__(XRRFreeOutputInfo) *p_XRRFreeOutputInfo =
        dlsym(RTLD_DEFAULT, "XRRFreeOutputInfo");
    __typeof__(XRRFreeCrtcInfo) *p_XRRFreeCrtcInfo =
        dlsym(RTLD_DEFAULT, "XRRFreeCrtcInfo");
    PRECOND(p_XRRFreeScreenResources);
    PRECOND(p_XRRFreeOutputInfo);
    PRECOND(p_XRRFreeCrtcInfo);

    Display *display = linux_x11_display();
    Window root = RootWindow(display, linux_x11_screen());
    XRRScreenResources *screen_resources =
        XRRGetScreenResources(display, root);
    ASSERT(screen_resources);

    XRROutputInfo *output_info = NULL;
    for (int i = 0; i < screen_resources->noutput; i++) {
        XRROutputInfo *oi = XRRGetOutputInfo(display, screen_resources,
                                             screen_resources->outputs[i]);
        if (oi) {
            if (oi->connection != RR_Disconnected) {
                output_info = oi;
                break;
            }
            (*p_XRRFreeOutputInfo)(oi);
        }
    }
    ASSERT(output_info);

    XRRCrtcInfo *crtc_info = XRRGetCrtcInfo(display, screen_resources,
                                            output_info->crtc);
    ASSERT(crtc_info);

    XRRModeInfo *mode = NULL;
    for (int i = 0; i < output_info->nmode; i++) {
        if (screen_resources->modes[i].id == crtc_info->mode) {
            mode = &screen_resources->modes[i];
            break;
        }
    }
    ASSERT(mode);

    *width_ret = mode->width;
    *height_ret = mode->height;
    (*p_XRRFreeCrtcInfo)(crtc_info);
    (*p_XRRFreeOutputInfo)(output_info);
    (*p_XRRFreeScreenResources)(screen_resources);
}

/*-----------------------------------------------------------------------*/

/**
 * glGetString_version_override:  Override function for glGetString().
 * When name is GL_VERSION, this function returns the string "1.0" instead
 * of the actual version to force opengl_init() to fail.
 */
static const GLubyte * GLAPIENTRY glGetString_version_override(
    GLenum name, TEST_glGetString_type *original_glGetString)
{
    if (name == GL_VERSION) {
        return (const GLubyte *)"1.0";
    }
    return (*original_glGetString)(name);
}

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_linux_graphics)

/*-----------------------------------------------------------------------*/

static void clear_variables(void)
{
    ASSERT(unsetenv("SIL_X11_CREATE_FULLSCREEN") == 0);
    ASSERT(unsetenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE") == 0);
    ASSERT(unsetenv("SIL_X11_EWMH_FULLSCREEN_RESIZE_AFTER") == 0);
    ASSERT(unsetenv("SIL_X11_FULLSCREEN_METHOD") == 0);
    ASSERT(unsetenv("SIL_X11_USE_TRANSIENT_FOR_HINT") == 0);
    ASSERT(unsetenv("SIL_X11_VIDEO_MODE_INTERFACE") == 0);
    ASSERT(unsetenv("SIL_X11_RESOURCE_CLASS") == 0);

    clear_x11_wrapper_variables();
}

/*-----------------------------------------------------------------------*/

TEST_INIT(init)
{
    clear_variables();
    if (!strstr(CURRENT_TEST_NAME(), "_NOINIT")) {
        CHECK_TRUE(graphics_init());
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST_CLEANUP(cleanup)
{
    if (!strstr(CURRENT_TEST_NAME(), "_NOCLEANUP")) {
        graphics_cleanup();
    }
    CHECK_INTEQUAL(linux_x11_get_error(), 0);
    clear_variables();
    return 1;
}

/*************************************************************************/
/******************** Test routines: Basic X11 logic *********************/
/*************************************************************************/

/* Note that we make heavy use of ASSERT() in some of these tests to ensure
 * that subsequent tests don't end up running in a broken environment. */

TEST(test_no_display_variable_NOINIT_NOCLEANUP)
{
    const char *saved_DISPLAY = getenv("DISPLAY");
    ASSERT(saved_DISPLAY);
    ASSERT(*saved_DISPLAY);

    linux_close_display();
    /* As long as we're here, check that the display pointer is cleared on
     * close to avoid any risk of an invalid pointer dereference (even
     * though the display will never be closed while the program is running
     * under normal circumstances). */
    CHECK_FALSE(linux_x11_display());

    ASSERT(unsetenv("DISPLAY") == 0);
    CHECK_FALSE(linux_open_display());

    ASSERT(setenv("DISPLAY", saved_DISPLAY, 1) == 0);
    ASSERT(linux_open_display());
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_empty_display_variable_NOINIT_NOCLEANUP)
{
    const char *saved_DISPLAY = getenv("DISPLAY");
    ASSERT(saved_DISPLAY);
    ASSERT(*saved_DISPLAY);

    linux_close_display();
    ASSERT(setenv("DISPLAY", "", 1) == 0);
    CHECK_FALSE(linux_open_display());

    ASSERT(setenv("DISPLAY", saved_DISPLAY, 1) == 0);
    ASSERT(linux_open_display());
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_bad_display_variable_NOINIT_NOCLEANUP)
{
    const char *saved_DISPLAY = getenv("DISPLAY");
    ASSERT(saved_DISPLAY);
    ASSERT(*saved_DISPLAY);

    linux_close_display();
    ASSERT(setenv("DISPLAY", ":-1", 1) == 0);
    CHECK_FALSE(linux_open_display());

    ASSERT(setenv("DISPLAY", saved_DISPLAY, 1) == 0);
    ASSERT(linux_open_display());
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_x11_error_handler)
{
    Display *display = linux_x11_display();

    XMapWindow(display, None);
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), BadWindow);
    const char *log = strstr(test_DLOG_last_message, ": ");
    ASSERT(log);
    log += 2;
    CHECK_STRSTARTS(log, "X11 error 3 (BadWindow");
    CHECK_STRENDS(log, ") in request 8 (X_MapWindow)");

    if (check_xrandr()) {
        __typeof__(XRRSetScreenSize) *p_XRRSetScreenSize =
            dlsym(RTLD_DEFAULT, "XRRSetScreenSize");
        ASSERT(p_XRRSetScreenSize != NULL);
        (*p_XRRSetScreenSize)(display, DefaultRootWindow(display), 0, 0, 0, 0);
        XSync(display, False);
        CHECK_INTEQUAL(linux_x11_get_error(), BadValue);
        log = strstr(test_DLOG_last_message, ": ");
        ASSERT(log);
        log += 2;
        CHECK_STRSTARTS(log, "X11 error 2 (BadValue");
        CHECK_STRENDS(log, ".7 (RRSetScreenSize)");
    }

    XGetAtomName(display, (Atom)-1);
    XSync(display, False);
    XMapWindow(display, None);  // Error should be discarded.
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), BadAtom);

    /* The saved error code should have been cleared by the previous call. */
    CHECK_INTEQUAL(linux_x11_get_error(), 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resource_class)
{
    ASSERT(!getenv("SIL_X11_RESOURCE_CLASS"));

    Display *display = linux_x11_display();
    Atom wm_class;
    CHECK_TRUE(wm_class = XInternAtom(display, "WM_CLASS", True));
    Atom utf8_string = XInternAtom(display, "UTF8_STRING", False);

    Window window;
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    window = linux_x11_window();
    result = XGetWindowProperty(
        display, window, wm_class, 0, 1000, False, AnyPropertyType,
        &actual_type, &actual_format, &nitems, &bytes_after, &prop);
    CHECK_INTEQUAL(result, Success);
    if (utf8_string && actual_type != utf8_string) {
        CHECK_INTEQUAL(actual_type, XA_STRING);
    }
    CHECK_INTEQUAL(actual_format, 8);
    CHECK_TRUE(nitems > 0);
    CHECK_TRUE(nitems > strlen((char *)prop) + 1);  // Should be 2 strings.
    XFree(prop);

    static const char test_class[] = "test_class";
    ASSERT(setenv("SIL_X11_RESOURCE_CLASS", test_class, 1) == 0);
    linux_close_window();
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    window = linux_x11_window();
    result = XGetWindowProperty(
        display, window, wm_class, 0, 1000, False, AnyPropertyType,
        &actual_type, &actual_format, &nitems, &bytes_after, &prop);
    CHECK_INTEQUAL(result, Success);
    /* The class name is an ASCII string so this should always be XA_STRING,
     * not UTF8_STRING. */
    CHECK_INTEQUAL(actual_type, XA_STRING);
    CHECK_INTEQUAL(actual_format, 8);
    CHECK_INTEQUAL(nitems, 2 * (strlen(test_class) + 1));
    CHECK_INTEQUAL(bytes_after, 0);
    CHECK_STREQUAL((char *)prop, test_class);
    CHECK_STREQUAL((char *)(prop + strlen(test_class) + 1), test_class);
    XFree(prop);

    /* An empty string should revert to the default class name. */
    ASSERT(setenv("SIL_X11_RESOURCE_CLASS", "", 1) == 0);
    linux_close_window();
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    window = linux_x11_window();
    result = XGetWindowProperty(
        display, window, wm_class, 0, 1000, False, AnyPropertyType,
        &actual_type, &actual_format, &nitems, &bytes_after, &prop);
    CHECK_INTEQUAL(result, Success);
    if (utf8_string && actual_type != utf8_string) {
        CHECK_INTEQUAL(actual_type, XA_STRING);
    }
    CHECK_INTEQUAL(actual_format, 8);
    CHECK_TRUE(nitems > 2);  // Should not be two empty strings.
    XFree(prop);

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Failure here could leave the window manager itself in an inconsistent
 * state, so be sure to clean up. */
#undef FAIL_ACTION
#define FAIL_ACTION  goto error

TEST(test_detect_window_manager)
{
    Display *display = linux_x11_display();
    Window root = RootWindow(display, 0);
    Atom motif_wm_info, net_supporting_wm_check, net_wm_name, utf8_string;
    Atom windowmaker_noticeboard;
    ASSERT(motif_wm_info = XInternAtom(display, "_MOTIF_WM_INFO", False));
    ASSERT(net_supporting_wm_check =
               XInternAtom(display, "_NET_SUPPORTING_WM_CHECK", False));
    ASSERT(net_wm_name = XInternAtom(display, "_NET_WM_NAME", False));
    ASSERT(utf8_string = XInternAtom(display, "UTF8_STRING", False));
    ASSERT(windowmaker_noticeboard =
               XInternAtom(display, "_WINDOWMAKER_NOTICEBOARD", False));

    Window supporting_window;
    Colormap colormap =
        XCreateColormap(display, root, DefaultVisual(display,0), AllocNone);
    XSetWindowAttributes cw_attributes = {
        .override_redirect = False,
        .background_pixmap = None,
        .border_pixel = 0,
        .colormap = colormap,
    };
    ASSERT(supporting_window = XCreateWindow(
               display, root, -1, -1, 1, 1, 0, DefaultDepth(display,0),
               InputOutput, DefaultVisual(display,0),
               CWOverrideRedirect | CWBackPixmap | CWBorderPixel | CWColormap,
               &cw_attributes));
    XSync(display, False);
    ASSERT(linux_x11_get_error() == 0);

    Window old_supporting_window = None;
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    int result = XGetWindowProperty(
        display, root, net_supporting_wm_check, 0, 1, False, XA_WINDOW,
        &actual_type, &actual_format, &nitems, &bytes_after, &prop);
    if (result == Success && actual_type != 0) {
        /* It's not beyond the realm of possibility that a strange WM could
         * set a non-Window type value or multiple Window values on this
         * property, but since our cleanup code only handles single Window
         * values, we abort in that case to avoid corrupting WM state. */
        ASSERT(actual_type == XA_WINDOW);
        ASSERT(actual_format == 32);
        ASSERT(nitems == 1);
        old_supporting_window = *(long *)(void *)prop;
        XFree(prop);
    }

    void *old_motif_wm_info = NULL;
    int old_motif_wm_info_len = 0;
    result = XGetWindowProperty(
        display, root, motif_wm_info, 0, 1000, False, motif_wm_info,
        &actual_type, &actual_format, &nitems, &bytes_after, &prop);
    if (result == Success && actual_type != 0) {
        ASSERT(actual_type == motif_wm_info);
        ASSERT(actual_format == 32);
        ASSERT(bytes_after == 0);
        old_motif_wm_info = prop;
        old_motif_wm_info_len = nitems;
    }

    /* Clear _NET_WM_SUPPORTING_CHECK and _MOTIF_WM_INFO and check that we
     * detect the lack of a window manager. */
    CHECK_TRUE(XDeleteProperty(display, root, net_supporting_wm_check));
    CHECK_TRUE(XDeleteProperty(display, root, motif_wm_info));
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), 0);
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNNAMED);

    /* Test MWM detection. */
    CHECK_TRUE(XChangeProperty(display, root, motif_wm_info, motif_wm_info, 32,
                               PropModeReplace, (void *)((long[]){2, 0}), 1));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_MWM);
    /* A wrong property type should cause detection to fail. */
    CHECK_TRUE(XChangeProperty(display, root, motif_wm_info, XA_STRING, 8,
                               PropModeReplace, (void *)"foo", 3));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNNAMED);
    /* Delete the property for subsequent tests. */
    CHECK_TRUE(XDeleteProperty(display, root, motif_wm_info));

    /* Install our dummy window as _NET_WM_SUPPORTING_CHECK for subsequent
     * tests.  Currently, our dummy supporting window has no _NET_WM_NAME
     * property, so we should be detected as an unnamed window manager. */
    CHECK_TRUE(XChangeProperty(display, root, net_supporting_wm_check,
                               XA_WINDOW, 32, PropModeReplace,
                               (void *)((long[]){supporting_window}), 1));
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), 0);
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNNAMED);

    /* Test Window Maker detection. */
    CHECK_TRUE(XChangeProperty(display, supporting_window,
                               windowmaker_noticeboard,
                               XA_WINDOW, 32, PropModeReplace,
                               (void *)((long[]){supporting_window}), 1));
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), 0);
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_WINDOWMAKER);
    /* An empty property should cause detection to fail. */
    CHECK_TRUE(XChangeProperty(display, supporting_window,
                               windowmaker_noticeboard,
                               XA_WINDOW, 32, PropModeReplace, NULL, 0));
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), 0);
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNNAMED);
    /* Delete the property for subsequent tests. */
    CHECK_TRUE(XDeleteProperty(display, supporting_window,
                               windowmaker_noticeboard));
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), 0);

    /* Test generic WM detection. */
    const char *name = "FVWM";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_FVWM);

    /* Test detection using XA_STRING instead of utf8_string. */
    name = "Fluxbox";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               XA_STRING, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_FLUXBOX);

    /* Test detection of an unknown name. */
    name = "<unknown name>";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNKNOWN);

    /* Test Enlightenment detection. */
    name = "e16";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_ENLIGHTENMENT);
    /* These should not be detected as Enlightenment. */
    name = "e";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNKNOWN);
    name = "e0";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNKNOWN);
    name = "e16 maybe?";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNKNOWN);

    /* Test IceWM detection. */
    name = "IceWM 0.0.0";
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               utf8_string, 8, PropModeReplace,
                               (void *)name, strlen(name)));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_ICEWM);

    /* Test handling of a wrong type in _NET_WM_NAME. */
    CHECK_TRUE(XChangeProperty(display, supporting_window, net_wm_name,
                               XA_WINDOW, 32, PropModeReplace,
                               (void *)((long[]){supporting_window}), 1));
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNNAMED);

    /* Test handling of a wrong type for _NET_WM_SUPPORTING_CHECK. */
    CHECK_TRUE(XChangeProperty(display, root, net_supporting_wm_check,
                               XA_STRING, 8, PropModeReplace,
                               (void *)"foo", 3));
    XSync(display, False);
    CHECK_INTEQUAL(linux_x11_get_error(), 0);
    linux_close_window();
    graphics_set_display_mode(TESTW, TESTH, NULL);
    CHECK_INTEQUAL(linux_window_manager(), WM_UNNAMED);

    /* Clean up after ourselves. */
    if (old_supporting_window) {
        ASSERT(XChangeProperty(display, root, net_supporting_wm_check,
                               XA_WINDOW, 32, PropModeReplace,
                               (void *)((long[]){old_supporting_window}), 1));
    } else {
        ASSERT(XDeleteProperty(display, root, net_supporting_wm_check));
    }
    XSync(display, False);
    ASSERT(linux_x11_get_error() == 0);
    return 1;

  error:
    if (old_supporting_window) {
        ASSERT(XChangeProperty(display, root, net_supporting_wm_check,
                               XA_WINDOW, 32, PropModeReplace,
                               (void *)((long[]){old_supporting_window}), 1));
    } else {
        ASSERT(XDeleteProperty(display, root, net_supporting_wm_check));
    }
    if (old_motif_wm_info) {
        ASSERT(XChangeProperty(display, root, motif_wm_info, motif_wm_info, 32,
                               PropModeReplace, old_motif_wm_info,
                               old_motif_wm_info_len));
        XFree(old_motif_wm_info);
    } else {
        ASSERT(XDeleteProperty(display, root, motif_wm_info));
    }
    XSync(display, False);
    ASSERT(linux_x11_get_error() == 0);
    return 0;
}

#undef FAIL_ACTION
#define FAIL_ACTION  return 0

/*************************************************************************/
/************** Test routines: Linux-specific window tests ***************/
/*************************************************************************/

TEST(test_create_blank_cursor_NOINIT)
{
    disable_XCreateBitmapFromData = 1;
    CHECK_TRUE(graphics_init());
    disable_XCreateBitmapFromData = 0;
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    graphics_show_mouse_pointer(0);
    CHECK_TRUE(graphics_get_mouse_pointer_state());

    graphics_cleanup();
    disable_XCreatePixmapCursor = 1;
    CHECK_TRUE(graphics_init());
    disable_XCreatePixmapCursor = 0;
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    graphics_show_mouse_pointer(0);
    CHECK_TRUE(graphics_get_mouse_pointer_state());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_title_memory_failure)
{
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    char *name;
    CHECK_MEMORY_FAILURES(
        (graphics_set_window_title("SIL Test"),
         name = NULL,
         XFetchName(linux_x11_display(), linux_x11_window(), &name),
         (name && strcmp(name, "SIL Test") == 0) || (XFree(name), 0)));
    XFree(name);
    name = NULL;
    XGetIconName(linux_x11_display(), linux_x11_window(), &name);
    CHECK_STREQUAL(name, "SIL Test");
    XFree(name);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_icon_memory_failure)
{
    Display *display = linux_x11_display();
    Atom net_wm_icon = XInternAtom(display, "_NET_WM_ICON", True);
    if (!net_wm_icon) {
        SKIP("_NET_WM_ICON atom not found.");
    }

    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    int texture;
    CHECK_TRUE(texture = texture_create(32, 32, 0, 0));
    uint8_t *pixels;
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            pixels[(y*32+x)*4+0] = x<<3;
            pixels[(y*32+x)*4+1] = y;
            pixels[(y*32+x)*4+2] = x<<3 ^ y;
            pixels[(y*32+x)*4+3] = (y*32+x)/4;
        }
    }
    texture_unlock(texture);

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    CHECK_MEMORY_FAILURES(
        (graphics_set_window_icon(texture),
         (XGetWindowProperty(display, linux_x11_window(), net_wm_icon,
                             0, 2+32*32, False, AnyPropertyType, &actual_type,
                             &actual_format, &nitems, &bytes_after, &prop)
          == Success && nitems == 2+32*32)));
    CHECK_INTEQUAL(actual_type, XA_CARDINAL);
    CHECK_INTEQUAL(actual_format, 32);
    CHECK_INTEQUAL(bytes_after, 0);
    const long *icon_data = (long *)(void *)prop;
    CHECK_INTEQUAL(icon_data[0], 32);
    CHECK_INTEQUAL(icon_data[1], 32);
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            const int r = x<<3, g = y, b = x<<3 ^ y, a = (y*32+x)/4;
            const long expected = a<<24 | r<<16 | g<<8 | b;
            if (icon_data[2+(y*32+x)] != expected) {
                FAIL("icon_data[%d] was 0x%08lX but should have been 0x%08lX",
                     2+(y*32+x), icon_data[2+(y*32+x)], expected);
            }
        }
    }
    XFree(prop);

    const uint8_t *pixels_ro;
    CHECK_TRUE(pixels_ro = texture_lock_readonly(texture));
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            CHECK_PIXEL(&pixels_ro[(y*32+x)*4],
                        x<<3, y, x<<3 ^ y, (y*32+x)/4, x, y);
        }
    }
    texture_destroy(texture);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_window_icon_early)
{
    Display *display = linux_x11_display();
    Atom net_wm_icon = XInternAtom(display, "_NET_WM_ICON", True);
    if (!net_wm_icon) {
        SKIP("_NET_WM_ICON atom not found.");
    }

    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    int texture;
    CHECK_TRUE(texture = texture_create(32, 32, 0, 0));
    uint8_t *pixels;
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            pixels[(y*32+x)*4+0] = x<<3;
            pixels[(y*32+x)*4+1] = y;
            pixels[(y*32+x)*4+2] = x<<3 ^ y;
            pixels[(y*32+x)*4+3] = (y*32+x)/4;
        }
    }
    texture_unlock(texture);
    graphics_set_window_icon(texture);
    texture_destroy(texture);

    linux_close_window();
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    CHECK_INTEQUAL(XGetWindowProperty(display, linux_x11_window(), net_wm_icon,
                                      0, 2+32*32, False, AnyPropertyType,
                                      &actual_type, &actual_format, &nitems,
                                      &bytes_after, &prop), Success);
    CHECK_INTEQUAL(actual_type, XA_CARDINAL);
    CHECK_INTEQUAL(actual_format, 32);
    CHECK_INTEQUAL(nitems, 2+32*32);
    CHECK_INTEQUAL(bytes_after, 0);
    const long *icon_data = (long *)(void *)prop;
    CHECK_INTEQUAL(icon_data[0], 32);
    CHECK_INTEQUAL(icon_data[1], 32);
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            const int r = x<<3, g = y, b = x<<3 ^ y, a = (y*32+x)/4;
            const long expected = a<<24 | r<<16 | g<<8 | b;
            if (icon_data[2+(y*32+x)] != expected) {
                FAIL("icon_data[%d] was 0x%08lX but should have been 0x%08lX",
                     2+(y*32+x), icon_data[2+(y*32+x)], expected);
            }
        }
    }
    XFree(prop);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XCreateColormap_failure)
{
    disable_XCreateColormap = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    disable_XCreateColormap = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XCreateWindow_failure)
{
    disable_XCreateWindow = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    disable_XCreateWindow = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XCreateWindow_error)
{
    error_XCreateWindow = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    error_XCreateWindow = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_glXQueryExtension_failure_NOINIT)
{
    disable_glXQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    disable_glXCreateWindow = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_OPENGL_NOT_FOUND);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_glXCreateWindow_failure)
{
    disable_glXCreateWindow = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    disable_glXCreateWindow = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_glXCreateNewContext_failure)
{
    disable_glXCreateNewContext = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    disable_glXCreateNewContext = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_glXMakeContextCurrent_failure)
{
    disable_glXMakeContextCurrent = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(TESTW, TESTH, &error));
    disable_glXMakeContextCurrent = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_gl_version_too_old)
{
    TEST_dyngl_override_glGetString(glGetString_version_override);
    GraphicsError error;
    const int result = graphics_set_display_mode(TESTW, TESTH, &error);
    TEST_dyngl_override_glGetString(NULL);
    CHECK_FALSE(result);
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_OPENGL_TOO_OLD);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resize_window_noncentered)
{
    CHECK_TRUE(graphics_set_display_attr("center_window", 0));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    called_XCreateWindow = 0;
    called_XMoveResizeWindow = 0;
    called_XResizeWindow = 0;
    CHECK_TRUE(graphics_set_display_mode(TESTW+64, TESTH+64, NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_INTEQUAL(called_XMoveResizeWindow, 0);
    CHECK_INTEQUAL(called_XResizeWindow, 1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resize_window_centered)
{
    CHECK_TRUE(graphics_set_display_attr("center_window", 1));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    called_XCreateWindow = 0;
    called_XMoveResizeWindow = 0;
    called_XResizeWindow = 0;
    CHECK_TRUE(graphics_set_display_mode(TESTW+64, TESTH-32, NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_INTEQUAL(called_XMoveResizeWindow, 1);
    CHECK_INTEQUAL(called_XResizeWindow, 0);
    /* Give the window manager a chance to handle the move request. */
    XSync(linux_x11_display(), False);
    nanosleep(&(struct timespec){0, 100*1000*1000}, NULL);
    int x1, y1;
    Window unused_child;
    XTranslateCoordinates(linux_x11_display(), linux_x11_window(),
                          RootWindow(linux_x11_display(), linux_x11_screen()),
                          0, 0, &x1, &y1, &unused_child);

    /* Check coordinates separately since window managers may place windows
     * slightly differently at creation time vs. with XMoveResize(). */
    called_XCreateWindow = 0;
    called_XMoveResizeWindow = 0;
    called_XResizeWindow = 0;
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_INTEQUAL(called_XMoveResizeWindow, 1);
    CHECK_INTEQUAL(called_XResizeWindow, 0);
    XSync(linux_x11_display(), False);
    nanosleep(&(struct timespec){0, 100*1000*1000}, NULL);
    int x2, y2;
    XTranslateCoordinates(linux_x11_display(), linux_x11_window(),
                          RootWindow(linux_x11_display(), linux_x11_screen()),
                          0, 0, &x2, &y2, &unused_child);
    CHECK_INTEQUAL(x2, x1+32);
    CHECK_INTEQUAL(y2, y1-16);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resize_window_error)
{
    CHECK_TRUE(graphics_set_display_attr("center_window", 1));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    called_XCreateWindow = 0;
    called_XMoveResizeWindow = 0;
    called_XResizeWindow = 0;
    error_XMoveResizeWindow = 1;
    CHECK_TRUE(graphics_set_display_mode(TESTW+64, TESTH+64, NULL));
    error_XMoveResizeWindow = 0;
    CHECK_INTEQUAL(called_XCreateWindow, 1);
    CHECK_INTEQUAL(called_XMoveResizeWindow, 1);
    CHECK_INTEQUAL(called_XResizeWindow, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resize_window_to_fullscreen)
{
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resize_window_from_fullscreen)
{
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_resize_window_different_gl_config)
{
    CHECK_TRUE(graphics_set_display_attr("depth_bits", 0));
    CHECK_TRUE(graphics_set_display_attr("stencil_bits", 0));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("depth_bits", 16));
    CHECK_TRUE(graphics_set_display_attr("stencil_bits", 8));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 1);

    return 1;
}

/*************************************************************************/
/***************** Test routines: Display mode handling ******************/
/*************************************************************************/

/* This doesn't test anything per se, but it logs a skip message if only
 * one screen is available so the user knows some code will not be
 * functionally tested. */
TEST(test_multiple_screens)
{
    if (graphics_num_devices() == 1) {
        SKIP("Only one display device is available; some tests in this"
             " file may spuriously pass.");
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_video_mode_interface_NOINIT)
{
    disable_XF86VidModeQueryExtension = 1;
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    disable_XRRQueryExtension = 0;

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (VidMode/XRandR disabled):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        /* All modes should have an unknown refresh rate. */
        CHECK_FLOATEQUAL(modes[i].refresh, 0);
        if (i >= 1) {
            /* There should only be one mode (the current mode) per screen. */
            CHECK_TRUE(modes[i].device == modes[i-1].device + 1);
        }
    }
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    /* graphics_frame_period() should return zero if we have no video mode
     * interface. */
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    CHECK_FLOATEQUAL(graphics_frame_period(), 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_video_mode_interface_memory_failure_NOINIT)
{
    disable_XF86VidModeQueryExtension = 1;
    disable_XRRQueryExtension = 1;
    CHECK_MEMORY_FAILURES(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    disable_XRRQueryExtension = 0;

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=NONE):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        if (i >= 1) {
            CHECK_TRUE(modes[i].device == modes[i-1].device + 1);
        }
    }
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_vidmode_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (XF86VidMode):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    int screen_1_mode = -1;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        if (modes[i].device > 0) {
            /* There should only be one mode (the current mode) for the
             * second and later screens. */
            if (modes[i].device == 1) {
                CHECK_TRUE(screen_1_mode == -1);
                screen_1_mode = i;
            }
            CHECK_TRUE(modes[i].device == modes[i-1].device + 1);
        }
    }
    if (screen_1_mode >= 0) {
        CHECK_INTEQUAL(mode_list->num_modes,
                       screen_1_mode + (graphics_num_devices() - 1));
    }
    CHECK_TRUE(called_XF86VidModeGetModeLine > 0);
    CHECK_TRUE(called_XF86VidModeGetAllModeLines > 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_vidmode_memory_failure_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    disable_XRRQueryExtension = 1;
    CHECK_MEMORY_FAILURES(graphics_init());
    disable_XRRQueryExtension = 0;

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    DLOG("Available display modes (XF86VidMode, memory failures):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);
    CHECK_TRUE(called_XF86VidModeGetModeLine > 0);
    CHECK_TRUE(called_XF86VidModeGetAllModeLines > 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XF86VidModeGetModeLine_failure_NOINIT)
{
    disable_XRRQueryExtension = 1;
    disable_XF86VidModeGetModeLine = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;
    disable_XF86VidModeGetModeLine = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_TRUE(called_XF86VidModeGetModeLine > 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XF86VidModeGetAllModeLines_failure_NOINIT)
{
    disable_XRRQueryExtension = 1;
    disable_XF86VidModeGetAllModeLines = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;
    disable_XF86VidModeGetAllModeLines = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_TRUE(called_XF86VidModeGetModeLine > 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines,
                   called_XF86VidModeGetModeLine);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_vidmode_fallback_memory_failure_NOINIT)
{
    disable_XRRQueryExtension = 1;
    disable_XF86VidModeGetModeLine = 1;
    CHECK_MEMORY_FAILURES(graphics_init());
    disable_XRRQueryExtension = 0;
    disable_XF86VidModeGetModeLine = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_TRUE(called_XF86VidModeGetModeLine > 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xrandr)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (XRandR):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    CHECK_TRUE(mode_list->num_modes > 0);
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
    }
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_TRUE(called_XRRGetOutputInfo > 0);
    CHECK_TRUE(called_XRRGetCrtcInfo > 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xrandr_memory_failure)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    CHECK_MEMORY_FAILURES(graphics_init());

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (XRandR, memory failures):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_TRUE(called_XRRGetOutputInfo > 0);
    CHECK_TRUE(called_XRRGetCrtcInfo > 0);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XRRQueryVersion_failure_NOINIT)
{
    disable_XF86VidModeQueryExtension = 1;
    disable_XRRQueryVersion = 1;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    disable_XRRQueryVersion = 0;

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xrandr_version_too_old_NOINIT)
{
    disable_XF86VidModeQueryExtension = 1;
    xrandr_version_major = 0;
    xrandr_version_minor = 9;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    xrandr_version_major = 0;
    xrandr_version_minor = 0;

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    graphics_cleanup();
    disable_XF86VidModeQueryExtension = 1;
    xrandr_version_major = 1;
    xrandr_version_minor = 1;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    xrandr_version_major = 0;
    xrandr_version_minor = 0;

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XRRGetScreenResources_failure_NOINIT)
{
    disable_XRRGetScreenResources = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRGetScreenResources = 0;

    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XRRGetOutputInfo_failure_NOINIT)
{
    disable_XRRGetOutputInfo = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRGetOutputInfo = 0;

    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_TRUE(called_XRRGetOutputInfo > 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_XRRGetCrtcInfo_failure_NOINIT)
{
    disable_XRRGetCrtcInfo = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRGetCrtcInfo = 0;

    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_TRUE(called_XRRGetOutputInfo > 0);
    CHECK_TRUE(called_XRRGetCrtcInfo > 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xrandr_fallback_memory_failure_NOINIT)
{
    disable_XRRGetCrtcInfo = 1;
    CHECK_MEMORY_FAILURES(graphics_init());
    disable_XRRGetCrtcInfo = 0;

    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_TRUE(called_XRRGetOutputInfo > 0);
    CHECK_TRUE(called_XRRGetCrtcInfo > 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_video_mode_none_NOINIT)
{
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "NONE", 1) == 0);
    CHECK_TRUE(graphics_init());

    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=NONE):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        if (i >= 1) {
            CHECK_TRUE(modes[i].device == modes[i-1].device + 1);
        }
    }
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_video_mode_vidmode_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "VIDMODE", 1) == 0);
    CHECK_TRUE(graphics_init());

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    DLOG("Available display modes (interface=VIDMODE):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);
    CHECK_TRUE(called_XF86VidModeGetModeLine > 0);
    CHECK_TRUE(called_XF86VidModeGetAllModeLines > 0);
    CHECK_INTEQUAL(called_XRRGetScreenResources, 0);
    CHECK_INTEQUAL(called_XRRGetOutputInfo, 0);
    CHECK_INTEQUAL(called_XRRGetCrtcInfo, 0);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_video_mode_xrandr)
{
    ASSERT(!getenv("SIL_X11_VIDEO_MODE_INTERFACE"));
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "XRANDR", 1) == 0);
    CHECK_TRUE(graphics_init());

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=XRANDR):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);
    CHECK_INTEQUAL(called_XF86VidModeGetModeLine, 0);
    CHECK_INTEQUAL(called_XF86VidModeGetAllModeLines, 0);
    CHECK_TRUE(called_XRRGetScreenResources > 0);
    CHECK_TRUE(called_XRRGetOutputInfo > 0);
    CHECK_TRUE(called_XRRGetCrtcInfo > 0);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_video_mode_empty)
{
    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "", 1) == 0);
    CHECK_TRUE(graphics_init());

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=<invalid>):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_video_mode_invalid)
{
    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    /* This (invalid) setting should be ignored. */
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "<invalid>", 1) == 0);
    CHECK_TRUE(graphics_init());

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=<invalid>):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Check that explicitly requesting XF86VidMode or XRandR doesn't enable
 * the interface if it doesn't exist on the system. */
TEST(test_env_video_mode_disabled_NOINIT)
{
    disable_XF86VidModeQueryExtension = 1;
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    disable_XRRQueryExtension = 0;

    GraphicsDisplayModeEntry *orig_modes;
    int orig_num_modes;
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    orig_num_modes = mode_list->num_modes;
    ASSERT(orig_modes = mem_alloc(sizeof(*orig_modes) * orig_num_modes, 0, 0));
    memcpy(orig_modes, mode_list->modes, sizeof(*orig_modes) * orig_num_modes);

    graphics_cleanup();
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "VIDMODE", 1) == 0);
    disable_XF86VidModeQueryExtension = 1;
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    disable_XRRQueryExtension = 0;

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=VIDMODE, disabled):");
    const GraphicsDisplayModeEntry *modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);

    graphics_cleanup();
    ASSERT(setenv("SIL_X11_VIDEO_MODE_INTERFACE", "XRANDR", 1) == 0);
    disable_XF86VidModeQueryExtension = 1;
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XF86VidModeQueryExtension = 0;
    disable_XRRQueryExtension = 0;

    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_TRUE(mode_list->num_modes > 0);
    DLOG("Available display modes (interface=XRANDR, disabled):");
    modes = mode_list->modes;
    for (int i = 0; i < mode_list->num_modes; i++) {
        DLOG("   Display %d (%s): %dx%d (%.4gHz)", modes[i].device,
             modes[i].device_name, modes[i].width, modes[i].height,
             modes[i].refresh);
        CHECK_INTEQUAL(modes[i].device, orig_modes[i].device);
        CHECK_INTEQUAL(modes[i].width, orig_modes[i].width);
        CHECK_INTEQUAL(modes[i].height, orig_modes[i].height);
        CHECK_FLOATEQUAL(modes[i].refresh, orig_modes[i].refresh);
    }
    CHECK_INTEQUAL(mode_list->num_modes, orig_num_modes);

    mem_free(orig_modes);
    return 1;
}

/*-----------------------------------------------------------------------*/

/* linux_reset_video_mode() is called by the abort handler, so make sure
 * it doesn't crash even if the graphics subsystem is not initialized. */
TEST(test_reset_video_mode_uninitted_NOINIT_NOCLEANUP)
{
    linux_reset_video_mode();
    return 1;
}

/*************************************************************************/
/**************** Test routines: Fullscreen mode changing ****************/
/*************************************************************************/

TEST(test_fullscreen_vidmode_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), mode_width);
    CHECK_INTEQUAL(graphics_display_height(), mode_height);
    CHECK_INTEQUAL(called_XRRSetCrtcConfig, 0);
    CHECK_INTEQUAL(called_XRRGetPanning, 0);

    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_xrandr_NOCLEANUP)
{
    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    if (!check_xrandr()) {
        graphics_cleanup();
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        graphics_cleanup();
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), mode_width);
    CHECK_INTEQUAL(graphics_display_height(), mode_height);
    CHECK_TRUE(called_XRRSetCrtcConfig > 0);
    CHECK_TRUE(called_XRRGetPanning > 0);

    graphics_cleanup();
    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_xrandr_memory_failure_NOCLEANUP)
{
    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    if (!check_xrandr()) {
        graphics_cleanup();
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        graphics_cleanup();
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_MEMORY_FAILURES(graphics_set_display_mode(mode_width, mode_height,
                                                    NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), mode_width);
    CHECK_INTEQUAL(graphics_display_height(), mode_height);
    CHECK_TRUE(called_XRRSetCrtcConfig > 0);
    CHECK_INTEQUAL(called_XRRGetPanning, 0);

    graphics_cleanup();
    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_xrandr_XRRGetScreenResources_failure)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    disable_XRRGetScreenResources = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(mode_width, mode_height, &error));
    disable_XRRGetScreenResources = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    CHECK_INTEQUAL(called_XRRSetCrtcConfig, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_xrandr_XRRGetOutputInfo_failure)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    disable_XRRGetOutputInfo = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(mode_width, mode_height, &error));
    disable_XRRGetOutputInfo = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    CHECK_INTEQUAL(called_XRRSetCrtcConfig, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_xrandr_XRRGetCrtcInfo_failure)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    disable_XRRGetCrtcInfo = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(mode_width, mode_height, &error));
    disable_XRRGetCrtcInfo = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    CHECK_INTEQUAL(called_XRRSetCrtcConfig, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_xrandr_v1_2_no_panning_NOINIT_NOCLEANUP)
{
    xrandr_version_major = 1;
    xrandr_version_minor = 2;
    CHECK_TRUE(graphics_init());
    xrandr_version_major = 0;
    xrandr_version_minor = 0;

    if (!check_xrandr()) {
        graphics_cleanup();
        SKIP("XRandR not found or too old.");
    }

    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        graphics_cleanup();
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), mode_width);
    CHECK_INTEQUAL(graphics_display_height(), mode_height);
    CHECK_INTRANGE(called_XRRSetCrtcConfig, 1, 2);
    CHECK_INTEQUAL(called_XRRGetPanning, 0);

    graphics_cleanup();
    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_switch_failure)
{
    if (!check_xrandr()) {
        graphics_cleanup();
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        graphics_cleanup();
        SKIP("No alternate video mode available.");
    }

    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    CHECK_TRUE(graphics_display_is_window());

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    disable_XRRGetCrtcInfo = 1;
    GraphicsError error;
    CHECK_FALSE(graphics_set_display_mode(mode_width, mode_height, &error));
    disable_XRRGetCrtcInfo = 0;
    CHECK_INTEQUAL(error, GRAPHICS_ERROR_MODE_NOT_SUPPORTED);
    /* The old window should still be open. */
    CHECK_TRUE(graphics_display_is_window());

    return 1;
}

/*************************************************************************/
/******************** Test routines: Early fullscreen ********************/
/*************************************************************************/

TEST(test_env_create_fullscreen_true)
{
    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);

    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();

    /* XMoveWindow() and XSendEvent() are only used for post-create
     * fullscreen. */
    CHECK_INTEQUAL(called_XMoveWindow, 0);
    CHECK_FALSE(last_event_display);

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result = BadMatch;  // Arbitrary non-"Success" value.

    if (net_wm_state && net_wm_state_fullscreen) {
        result = XGetWindowProperty(display, window, net_wm_state, 0, 1,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        if (result == Success && actual_type == 0) {
            result = BadMatch;
        }
    }
    if (result == Success) {
        CHECK_INTEQUAL(actual_type, XA_ATOM);
        CHECK_INTEQUAL(actual_format, 32);
        CHECK_INTEQUAL(nitems, 1);
        CHECK_INTEQUAL(bytes_after, 0);
        CHECK_INTEQUAL(*(long *)(void *)prop, net_wm_state_fullscreen);
        XFree(prop);
    } else {
        if (motif_wm_hints) {
            result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                        False, AnyPropertyType, &actual_type,
                                        &actual_format, &nitems, &bytes_after,
                                        &prop);
            CHECK_INTEQUAL(result, Success);
            CHECK_INTEQUAL(actual_type, motif_wm_hints);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 5);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
            CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
            XFree(prop);
        } else {
            result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR,
                                        0, 1, False, AnyPropertyType,
                                        &actual_type, &actual_format, &nitems,
                                        &bytes_after, &prop);
            CHECK_INTEQUAL(result, Success);
            CHECK_INTEQUAL(actual_type, XA_WINDOW);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 1);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(*(long *)(void *)prop,
                           RootWindow(display, linux_x11_screen()));
            XFree(prop);
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_create_fullscreen_false)
{
    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    const long net_wm_state_add = 1;  // Not an atom, despite the spelling.
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);

    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "0", 1) == 0);
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();

    if (last_event_display) {
        CHECK_TRUE(net_wm_state);
        CHECK_TRUE(net_wm_state_fullscreen);
        Window root = RootWindow(display, linux_x11_screen());
        CHECK_PTREQUAL(last_event_display, display);
        CHECK_INTEQUAL(last_event_window, root);
        CHECK_FALSE(last_event_propagate);
        CHECK_INTEQUAL(last_event_mask,
                       SubstructureNotifyMask | SubstructureRedirectMask);
        CHECK_INTEQUAL(last_event.type, ClientMessage);
        CHECK_PTREQUAL(last_event.xclient.display, display);
        CHECK_INTEQUAL(last_event.xclient.window, window);
        CHECK_INTEQUAL(last_event.xclient.message_type, net_wm_state);
        CHECK_INTEQUAL(last_event.xclient.format, 32);
        CHECK_INTEQUAL(last_event.xclient.data.l[0], net_wm_state_add);
        CHECK_INTEQUAL(last_event.xclient.data.l[1], net_wm_state_fullscreen);
        CHECK_INTEQUAL(last_event.xclient.data.l[2], 0);
        CHECK_INTEQUAL(last_event.xclient.data.l[3], 1);
    } else {
        CHECK_INTEQUAL(called_XMoveWindow, 1);
        Atom actual_type;
        int actual_format;
        unsigned long nitems, bytes_after;
        unsigned char *prop;
        Status result;
        if (motif_wm_hints) {
            result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                        False, AnyPropertyType, &actual_type,
                                        &actual_format, &nitems, &bytes_after,
                                        &prop);
            CHECK_INTEQUAL(result, Success);
            CHECK_INTEQUAL(actual_type, motif_wm_hints);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 5);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
            CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
            XFree(prop);
        } else {
            result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR,
                                        0, 1, False, AnyPropertyType,
                                        &actual_type, &actual_format, &nitems,
                                        &bytes_after, &prop);
            CHECK_INTEQUAL(result, Success);
            CHECK_INTEQUAL(actual_type, XA_WINDOW);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 1);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(*(long *)(void *)prop,
                           RootWindow(display, linux_x11_screen()));
            XFree(prop);
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

/* Both invalid and empty values for $SIL_CREATE_FULLSCREEN should have the
 * same effect (and the logic to test them is similarly identical), so we
 * do them both here in a loop. */
TEST(test_env_create_fullscreen_invalid_empty)
{
    ASSERT(!getenv("SIL_CREATE_FULLSCREEN"));

    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    const long net_wm_state_add = 1;
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);

    /* Figure out the default fullscreen timing and method. */
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    const int is_early_fullscreen = !called_XMoveWindow && !last_event_display;
    int is_ewmh_fullscreen = 0;
    if (last_event_display) {
        CHECK_TRUE(net_wm_state);
        CHECK_TRUE(net_wm_state_fullscreen);
        is_ewmh_fullscreen = 1;
    } else if (net_wm_state && net_wm_state_fullscreen) {
        Window window = linux_x11_window();
        Atom actual_type;
        int actual_format;
        unsigned long nitems, bytes_after;
        unsigned char *prop;
        Status result;
        result = XGetWindowProperty(display, window, net_wm_state, 0, 1,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        if (result == Success && actual_type != 0) {
            is_ewmh_fullscreen = 1;
            XFree(prop);
        }
    }

    for (int do_empty = 0; do_empty < 2; do_empty++) {
        /* This setting should be ignored. */
        ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN",
                      do_empty ? "" : "foo", 1) == 0);
        linux_close_window();
        CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                             graphics_device_height(), NULL));
        Window window = linux_x11_window();

        if (is_ewmh_fullscreen) {

            CHECK_TRUE(net_wm_state);
            CHECK_TRUE(net_wm_state_fullscreen);
            if (is_early_fullscreen) {
                CHECK_FALSE(last_event_display);
                Atom actual_type;
                int actual_format;
                unsigned long nitems, bytes_after;
                unsigned char *prop;
                Status result = XGetWindowProperty(
                    display, window, net_wm_state, 0, 1, False,
                    AnyPropertyType, &actual_type, &actual_format,
                    &nitems, &bytes_after, &prop);
                CHECK_INTEQUAL(result, Success);
                CHECK_INTEQUAL(actual_type, XA_ATOM);
                CHECK_INTEQUAL(actual_format, 32);
                CHECK_INTEQUAL(nitems, 1);
                CHECK_INTEQUAL(bytes_after, 0);
                CHECK_INTEQUAL(*(long *)(void *)prop, net_wm_state_fullscreen);
                XFree(prop);
            } else {
                Window root = RootWindow(display, linux_x11_screen());
                CHECK_PTREQUAL(last_event_display, display);
                CHECK_INTEQUAL(last_event_window, root);
                CHECK_FALSE(last_event_propagate);
                CHECK_INTEQUAL(last_event_mask,
                               SubstructureNotifyMask | SubstructureRedirectMask);
                CHECK_INTEQUAL(last_event.type, ClientMessage);
                CHECK_PTREQUAL(last_event.xclient.display, display);
                CHECK_INTEQUAL(last_event.xclient.window, window);
                CHECK_INTEQUAL(last_event.xclient.message_type, net_wm_state);
                CHECK_INTEQUAL(last_event.xclient.format, 32);
                CHECK_INTEQUAL(last_event.xclient.data.l[0], net_wm_state_add);
                CHECK_INTEQUAL(last_event.xclient.data.l[1], net_wm_state_fullscreen);
                CHECK_INTEQUAL(last_event.xclient.data.l[2], 0);
                CHECK_INTEQUAL(last_event.xclient.data.l[3], 1);
            }

        } else {  // !is_ewmh_fullscreen

            CHECK_INTEQUAL(called_XMoveWindow, is_early_fullscreen ? 0 : 1);
            Atom actual_type;
            int actual_format;
            unsigned long nitems, bytes_after;
            unsigned char *prop;
            Status result;
            if (motif_wm_hints) {
                result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                            False, AnyPropertyType, &actual_type,
                                            &actual_format, &nitems, &bytes_after,
                                            &prop);
                CHECK_INTEQUAL(result, Success);
                CHECK_INTEQUAL(actual_type, motif_wm_hints);
                CHECK_INTEQUAL(actual_format, 32);
                CHECK_INTEQUAL(nitems, 5);
                CHECK_INTEQUAL(bytes_after, 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
                CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
                XFree(prop);
            } else {
                result = XGetWindowProperty(
                    display, window, XA_WM_TRANSIENT_FOR, 0, 1, False,
                    AnyPropertyType, &actual_type, &actual_format, &nitems,
                    &bytes_after, &prop);
                CHECK_INTEQUAL(result, Success);
                CHECK_INTEQUAL(actual_type, XA_WINDOW);
                CHECK_INTEQUAL(actual_format, 32);
                CHECK_INTEQUAL(nitems, 1);
                CHECK_INTEQUAL(bytes_after, 0);
                CHECK_INTEQUAL(*(long *)(void *)prop,
                               RootWindow(display, linux_x11_screen()));
                XFree(prop);
            }

        }  // if (is_ewmh_fullscreen)
    }  // for (int do_empty = 0; do_empty < 2; do_empty++)

    return 1;
}

/*************************************************************************/
/******************* Test routines: Fullscreen methods *******************/
/*************************************************************************/

TEST(test_env_fullscreen_method_XMoveWindow)
{
    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "XMOVEWINDOW", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    CHECK_INTEQUAL(called_XMoveWindow, 0);
    if (motif_wm_hints) {
        result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_INTEQUAL(result, Success);
        CHECK_INTEQUAL(actual_type, motif_wm_hints);
        CHECK_INTEQUAL(actual_format, 32);
        CHECK_INTEQUAL(nitems, 5);
        CHECK_INTEQUAL(bytes_after, 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
        CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
        XFree(prop);
    } else {
        result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_INTEQUAL(result, Success);
        CHECK_INTEQUAL(actual_type, XA_WINDOW);
        CHECK_INTEQUAL(actual_format, 32);
        CHECK_INTEQUAL(nitems, 1);
        CHECK_INTEQUAL(bytes_after, 0);
        CHECK_INTEQUAL(*(long *)(void *)prop,
                       RootWindow(display, linux_x11_screen()));
        XFree(prop);
    }

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_INTEQUAL(called_XMoveWindow, 1);
    if (motif_wm_hints) {
        result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_INTEQUAL(result, Success);
        CHECK_INTEQUAL(actual_type, motif_wm_hints);
        CHECK_INTEQUAL(actual_format, 32);
        CHECK_INTEQUAL(nitems, 5);
        CHECK_INTEQUAL(bytes_after, 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
        CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[2], 1);
        CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
        XFree(prop);
    } else {
        result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_TRUE(result != Success || actual_type == 0 || *(long *)(void *)prop == 0);
    }

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_INTEQUAL(called_XMoveWindow, 2);
    if (motif_wm_hints) {
        result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_INTEQUAL(result, Success);
        CHECK_INTEQUAL(actual_type, motif_wm_hints);
        CHECK_INTEQUAL(actual_format, 32);
        CHECK_INTEQUAL(nitems, 5);
        CHECK_INTEQUAL(bytes_after, 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
        CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
        CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
        XFree(prop);
    } else {
        result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_INTEQUAL(result, Success);
        CHECK_INTEQUAL(actual_type, XA_WINDOW);
        CHECK_INTEQUAL(actual_format, 32);
        CHECK_INTEQUAL(nitems, 1);
        CHECK_INTEQUAL(bytes_after, 0);
        CHECK_INTEQUAL(*(long *)(void *)prop,
                       RootWindow(display, linux_x11_screen()));
        XFree(prop);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_fullscreen_method_ewmh_fullscreen)
{
    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    const long net_wm_state_add = 1;
    const long net_wm_state_remove = 0;
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    Window root = RootWindow(display, linux_x11_screen());
    result = XGetWindowProperty(display, window, net_wm_state, 0, 1, False,
                                AnyPropertyType, &actual_type, &actual_format,
                                &nitems, &bytes_after, &prop);
    CHECK_TRUE(result == Success);
    CHECK_INTEQUAL(actual_type, XA_ATOM);
    CHECK_INTEQUAL(actual_format, 32);
    CHECK_INTEQUAL(nitems, 1);
    CHECK_INTEQUAL(bytes_after, 0);
    CHECK_INTEQUAL(*(long *)(void *)prop, net_wm_state_fullscreen);
    XFree(prop);
    if (motif_wm_hints) {
        result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_TRUE(result != Success || actual_type == 0);
    }
    result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_TRUE(result != Success || actual_type == 0);

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_PTREQUAL(last_event_display, display);
    CHECK_INTEQUAL(last_event_window, root);
    CHECK_FALSE(last_event_propagate);
    CHECK_INTEQUAL(last_event_mask,
                   SubstructureNotifyMask | SubstructureRedirectMask);
    CHECK_INTEQUAL(last_event.type, ClientMessage);
    CHECK_PTREQUAL(last_event.xclient.display, display);
    CHECK_INTEQUAL(last_event.xclient.window, window);
    CHECK_INTEQUAL(last_event.xclient.message_type, net_wm_state);
    CHECK_INTEQUAL(last_event.xclient.format, 32);
    CHECK_INTEQUAL(last_event.xclient.data.l[0], net_wm_state_remove);
    CHECK_INTEQUAL(last_event.xclient.data.l[1], net_wm_state_fullscreen);
    CHECK_INTEQUAL(last_event.xclient.data.l[2], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[3], 1);
    if (motif_wm_hints) {
        result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_TRUE(result != Success || actual_type == 0);
    }
    result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_TRUE(result != Success || actual_type == 0);

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_PTREQUAL(last_event_display, display);
    CHECK_INTEQUAL(last_event_window, root);
    CHECK_FALSE(last_event_propagate);
    CHECK_INTEQUAL(last_event_mask,
                   SubstructureNotifyMask | SubstructureRedirectMask);
    CHECK_INTEQUAL(last_event.type, ClientMessage);
    CHECK_PTREQUAL(last_event.xclient.display, display);
    CHECK_INTEQUAL(last_event.xclient.window, window);
    CHECK_INTEQUAL(last_event.xclient.message_type, net_wm_state);
    CHECK_INTEQUAL(last_event.xclient.format, 32);
    CHECK_INTEQUAL(last_event.xclient.data.l[0], net_wm_state_add);
    CHECK_INTEQUAL(last_event.xclient.data.l[1], net_wm_state_fullscreen);
    CHECK_INTEQUAL(last_event.xclient.data.l[2], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[3], 1);
    if (motif_wm_hints) {
        result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        CHECK_TRUE(result != Success || actual_type == 0);
    }
    result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_TRUE(result != Success || actual_type == 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_fullscreen_method_invalid_empty)
{
    ASSERT(!getenv("SIL_X11_FULLSCREEN_METHOD"));

    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    int is_ewmh_fullscreen = 0;
    if (net_wm_state && net_wm_state_fullscreen) {
        Window window = linux_x11_window();
        result = XGetWindowProperty(display, window, net_wm_state, 0, 1,
                                    False, AnyPropertyType, &actual_type,
                                    &actual_format, &nitems, &bytes_after,
                                    &prop);
        if (result == Success && actual_type != 0) {
            is_ewmh_fullscreen = 1;
            XFree(prop);
        }
    }

    for (int do_empty = 0; do_empty < 2; do_empty++) {
        ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD",
                      do_empty ? "" : "foo", 1) == 0);

        linux_close_window();
        CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                             graphics_device_height(), NULL));
        Window window = linux_x11_window();
        if (is_ewmh_fullscreen) {
            result = XGetWindowProperty(display, window, net_wm_state, 0, 1,
                                        False, AnyPropertyType, &actual_type,
                                        &actual_format, &nitems, &bytes_after,
                                        &prop);
            CHECK_TRUE(result == Success);
            CHECK_INTEQUAL(actual_type, XA_ATOM);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 1);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(*(long *)(void *)prop, net_wm_state_fullscreen);
            XFree(prop);
            if (motif_wm_hints) {
                result = XGetWindowProperty(display, window, motif_wm_hints,
                                            0, 5, False, AnyPropertyType,
                                            &actual_type, &actual_format,
                                            &nitems, &bytes_after, &prop);
                CHECK_TRUE(result != Success || actual_type == 0);
            }
            result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR,
                                        0, 1, False, AnyPropertyType,
                                        &actual_type, &actual_format, &nitems,
                                        &bytes_after, &prop);
            CHECK_TRUE(result != Success || actual_type == 0);
        } else {  // !ewmh_fullscreen
            if (motif_wm_hints) {
                result = XGetWindowProperty(display, window, motif_wm_hints,
                                            0, 5, False, AnyPropertyType,
                                            &actual_type, &actual_format,
                                            &nitems, &bytes_after, &prop);
                CHECK_INTEQUAL(result, Success);
                CHECK_INTEQUAL(actual_type, motif_wm_hints);
                CHECK_INTEQUAL(actual_format, 32);
                CHECK_INTEQUAL(nitems, 5);
                CHECK_INTEQUAL(bytes_after, 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
                CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
                CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
            } else {
                result = XGetWindowProperty(
                    display, window, XA_WM_TRANSIENT_FOR, 0, 1, False,
                    AnyPropertyType, &actual_type, &actual_format, &nitems,
                    &bytes_after, &prop);
                CHECK_INTEQUAL(result, Success);
                CHECK_INTEQUAL(actual_type, XA_WINDOW);
                CHECK_INTEQUAL(actual_format, 32);
                CHECK_INTEQUAL(nitems, 1);
                CHECK_INTEQUAL(bytes_after, 0);
                CHECK_INTEQUAL(*(long *)(void *)prop,
                               RootWindow(display, linux_x11_screen()));
            }
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_move_before_true_vidmode_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    Display *display = linux_x11_display();
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    /* We set CREATE_FULLSCREEN=1 for these tests to avoid a _NET_WM_STATE
     * message being sent to the root window and overwriting the saved
     * event (or lack thereof) from setting _NET_WM_FULLSCREEN_MONITORS. */
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XMoveWindow, 1);
    CHECK_FALSE(last_event_display);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_move_before_true_xrandr)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found.");
    }

    if (graphics_num_devices() == 1) {
        SKIP("Only one display device present.");
    }

    Display *display = linux_x11_display();
    if (ScreenCount(display) > 1) {
        SKIP("X11 server not in a single-screen, multi-head configuration.");
    }

    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("device", 1));
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XMoveWindow, 1);
    CHECK_FALSE(last_event_display);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_move_before_false_vidmode_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    Display *display = linux_x11_display();
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    Atom net_wm_fullscreen_monitors =
        XInternAtom(display, "_NET_WM_FULLSCREEN_MONITORS", True);
    if (!net_wm_state || !net_wm_state_fullscreen
     || !net_wm_fullscreen_monitors) {
        SKIP("_NET_WM_STATE, _NET_WM_STATE_FULLSCREEN, or"
             " _NET_WM_FULLSCREEN_MONITORS atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE", "0", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    Window root = RootWindow(display, linux_x11_screen());
    CHECK_INTEQUAL(called_XMoveWindow, 0);
    CHECK_PTREQUAL(last_event_display, display);
    CHECK_INTEQUAL(last_event_window, root);
    CHECK_FALSE(last_event_propagate);
    CHECK_INTEQUAL(last_event_mask,
                   SubstructureNotifyMask | SubstructureRedirectMask);
    CHECK_INTEQUAL(last_event.type, ClientMessage);
    CHECK_PTREQUAL(last_event.xclient.display, display);
    CHECK_INTEQUAL(last_event.xclient.window, window);
    CHECK_INTEQUAL(last_event.xclient.message_type,
                   net_wm_fullscreen_monitors);
    CHECK_INTEQUAL(last_event.xclient.format, 32);
    CHECK_INTEQUAL(last_event.xclient.data.l[0], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[1], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[2], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[3], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[4], 1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_move_before_false_xrandr)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found.");
    }

    if (graphics_num_devices() == 1) {
        SKIP("Only one display device present.");
    }

    Display *display = linux_x11_display();
    if (ScreenCount(display) > 1) {
        SKIP("X11 server not in a single-screen, multi-head configuration.");
    }

    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    Atom net_wm_fullscreen_monitors =
        XInternAtom(display, "_NET_WM_FULLSCREEN_MONITORS", True);
    if (!net_wm_state || !net_wm_state_fullscreen
     || !net_wm_fullscreen_monitors) {
        SKIP("_NET_WM_STATE, _NET_WM_STATE_FULLSCREEN, or"
             " _NET_WM_FULLSCREEN_MONITORS atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE", "0", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("device", 1));
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    Window root = RootWindow(display, linux_x11_screen());
    CHECK_INTEQUAL(called_XMoveWindow, 0);
    CHECK_PTREQUAL(last_event_display, display);
    CHECK_INTEQUAL(last_event_window, root);
    CHECK_FALSE(last_event_propagate);
    CHECK_INTEQUAL(last_event_mask,
                   SubstructureNotifyMask | SubstructureRedirectMask);
    CHECK_INTEQUAL(last_event.type, ClientMessage);
    CHECK_PTREQUAL(last_event.xclient.display, display);
    CHECK_INTEQUAL(last_event.xclient.window, window);
    CHECK_INTEQUAL(last_event.xclient.message_type,
                   net_wm_fullscreen_monitors);
    CHECK_INTEQUAL(last_event.xclient.format, 32);
    CHECK_INTEQUAL(last_event.xclient.data.l[0], 1);
    CHECK_INTEQUAL(last_event.xclient.data.l[1], 1);
    CHECK_INTEQUAL(last_event.xclient.data.l[2], 1);
    CHECK_INTEQUAL(last_event.xclient.data.l[3], 1);
    CHECK_INTEQUAL(last_event.xclient.data.l[4], 1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_move_before_invalid_empty)
{
    Display *display = linux_x11_display();
    int device;
    if (check_xrandr()
     && graphics_num_devices() > 1
     && ScreenCount(display) == 1) {
        device = 1;
    } else {
        graphics_cleanup();
        disable_XRRQueryExtension = 1;
        CHECK_TRUE(graphics_init());
        disable_XRRQueryExtension = 0;
        if (!check_vidmode()) {
            SKIP("XF86VidMode not found.");
        }
        device = 0;
    }

    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    Atom net_wm_fullscreen_monitors =
        XInternAtom(display, "_NET_WM_FULLSCREEN_MONITORS", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);
    ASSERT(!getenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE"));

    CHECK_TRUE(graphics_set_display_attr("device", device));
    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    const int is_move_before = (called_XMoveWindow == 1);

    for (int do_empty = 0; do_empty < 2; do_empty++) {
        ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_MOVE_BEFORE",
                      do_empty ? "" : "foo", 1) == 0);

        called_XMoveWindow = 0;
        last_event_display = NULL;
        linux_close_window();
        CHECK_TRUE(graphics_set_display_attr("device", device));
        CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                             graphics_device_height(), NULL));
        if (is_move_before) {
            CHECK_INTEQUAL(called_XMoveWindow, 1);
        } else {
            CHECK_TRUE(net_wm_fullscreen_monitors);
            Window window = linux_x11_window();
            Window root = RootWindow(display, linux_x11_screen());
            CHECK_INTEQUAL(called_XMoveWindow, 0);
            CHECK_PTREQUAL(last_event_display, display);
            CHECK_INTEQUAL(last_event_window, root);
            CHECK_FALSE(last_event_propagate);
            CHECK_INTEQUAL(last_event_mask,
                           SubstructureNotifyMask | SubstructureRedirectMask);
            CHECK_INTEQUAL(last_event.type, ClientMessage);
            CHECK_PTREQUAL(last_event.xclient.display, display);
            CHECK_INTEQUAL(last_event.xclient.window, window);
            CHECK_INTEQUAL(last_event.xclient.message_type,
                           net_wm_fullscreen_monitors);
            CHECK_INTEQUAL(last_event.xclient.format, 32);
            CHECK_INTEQUAL(last_event.xclient.data.l[0], 1);
            CHECK_INTEQUAL(last_event.xclient.data.l[1], 1);
            CHECK_INTEQUAL(last_event.xclient.data.l[2], 1);
            CHECK_INTEQUAL(last_event.xclient.data.l[3], 1);
            CHECK_INTEQUAL(last_event.xclient.data.l[4], 1);
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_resize_after_true)
{
    Display *display = linux_x11_display();
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_RESIZE_AFTER", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    CHECK_INTEQUAL(called_XMoveResizeWindow, 1);
    XSizeHints hints;
    CHECK_TRUE(XGetWMNormalHints(display, window, &hints, (long[1]){}));
    CHECK_TRUE(hints.flags & PMinSize);
    CHECK_TRUE(hints.flags & PMaxSize);
    CHECK_INTEQUAL(hints.min_width, graphics_device_width());
    CHECK_INTEQUAL(hints.max_width, graphics_device_width());
    CHECK_INTEQUAL(hints.min_height, graphics_device_height());
    CHECK_INTEQUAL(hints.max_height, graphics_device_height());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_resize_after_false)
{
    Display *display = linux_x11_display();
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_RESIZE_AFTER", "0", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    CHECK_INTEQUAL(called_XMoveResizeWindow, 0);
    XSizeHints hints;
    CHECK_TRUE(XGetWMNormalHints(display, window, &hints, (long[1]){}));
    CHECK_FALSE(hints.flags & (PMinSize | PMaxSize));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_resize_after_invalid_empty)
{
    Display *display = linux_x11_display();
    Atom net_wm_state = XInternAtom(display, "_NET_WM_STATE", True);
    Atom net_wm_state_fullscreen =
        XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", True);
    if (!net_wm_state || !net_wm_state_fullscreen) {
        SKIP("_NET_WM_STATE or _NET_WM_STATE_FULLSCREEN atom not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "EWMH_FULLSCREEN", 1) == 0);
    ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_RESIZE_AFTER", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    const int is_resize_after = (called_XMoveResizeWindow == 1);

    for (int do_empty = 0; do_empty < 2; do_empty++) {
        ASSERT(setenv("SIL_X11_EWMH_FULLSCREEN_RESIZE_AFTER",
                      do_empty ? "" : "foo", 1) == 0);

        called_XMoveResizeWindow = 0;
        linux_close_window();
        CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                             graphics_device_height(), NULL));
        if (is_resize_after) {
            CHECK_INTEQUAL(called_XMoveResizeWindow, 1);
        } else {
            Window window = linux_x11_window();
            XSizeHints hints;
            CHECK_TRUE(XGetWMNormalHints(display, window, &hints, (long[1]){}));
            CHECK_TRUE(hints.flags & PMinSize);
            CHECK_TRUE(hints.flags & PMaxSize);
            CHECK_INTEQUAL(hints.min_width, graphics_device_width());
            CHECK_INTEQUAL(hints.max_width, graphics_device_width());
            CHECK_INTEQUAL(hints.min_height, graphics_device_height());
            CHECK_INTEQUAL(hints.max_height, graphics_device_height());
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_use_transient_for_true)
{
    Display *display = linux_x11_display();

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "XMOVEWINDOW", 1) == 0);
    ASSERT(setenv("SIL_X11_USE_TRANSIENT_FOR_HINT", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_INTEQUAL(result, Success);
    CHECK_INTEQUAL(actual_type, XA_WINDOW);
    CHECK_INTEQUAL(actual_format, 32);
    CHECK_INTEQUAL(nitems, 1);
    CHECK_INTEQUAL(bytes_after, 0);
    CHECK_INTEQUAL(*(long *)(void *)prop, RootWindow(display, linux_x11_screen()));
    XFree(prop);

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR, 0, 1,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_TRUE(result != Success || actual_type == 0 || *(long *)(void *)prop == 0);
    if (result == Success && actual_type != 0) {
        XFree(prop);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_use_transient_for_false)
{
    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);
    if (!motif_wm_hints) {
        SKIP("_MOTIF_WM_HINTS atom not found.");
    }

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "XMOVEWINDOW", 1) == 0);
    ASSERT(setenv("SIL_X11_USE_TRANSIENT_FOR_HINT", "0", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    Window window = linux_x11_window();
    result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_INTEQUAL(result, Success);
    CHECK_INTEQUAL(actual_type, motif_wm_hints);
    CHECK_INTEQUAL(actual_format, 32);
    CHECK_INTEQUAL(nitems, 5);
    CHECK_INTEQUAL(bytes_after, 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
    CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
    XFree(prop);

    called_XCreateWindow = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after, &prop);
    CHECK_INTEQUAL(result, Success);
    CHECK_INTEQUAL(actual_type, motif_wm_hints);
    CHECK_INTEQUAL(actual_format, 32);
    CHECK_INTEQUAL(nitems, 5);
    CHECK_INTEQUAL(bytes_after, 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
    CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[2], 1);
    CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
    CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
    XFree(prop);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_env_use_transient_for_invalid_empty)
{
    Display *display = linux_x11_display();
    Atom motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", True);

    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *prop;
    Status result;

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "XMOVEWINDOW", 1) == 0);

    for (int do_empty = 0; do_empty < 2; do_empty++) {
        ASSERT(setenv("SIL_X11_USE_TRANSIENT_FOR_HINT",
                      do_empty ? "" : "foo", 1) == 0);

        linux_close_window();
        CHECK_TRUE(graphics_set_display_attr("window", 0));
        CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                             graphics_device_height(), NULL));
        Window window = linux_x11_window();

        if (motif_wm_hints) {
            result = XGetWindowProperty(display, window, motif_wm_hints, 0, 5,
                                        False, AnyPropertyType, &actual_type,
                                        &actual_format, &nitems, &bytes_after,
                                        &prop);
            CHECK_INTEQUAL(result, Success);
            CHECK_INTEQUAL(actual_type, motif_wm_hints);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 5);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[0], 1 << 1);
            CHECK_INTEQUAL(((long *)(void *)prop)[1], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[2], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[3], 0);
            CHECK_INTEQUAL(((long *)(void *)prop)[4], 0);
            XFree(prop);
        } else {
            result = XGetWindowProperty(display, window, XA_WM_TRANSIENT_FOR,
                                        0, 1, False, AnyPropertyType,
                                        &actual_type, &actual_format, &nitems,
                                        &bytes_after, &prop);
            CHECK_INTEQUAL(result, Success);
            CHECK_INTEQUAL(actual_type, XA_WINDOW);
            CHECK_INTEQUAL(actual_format, 32);
            CHECK_INTEQUAL(nitems, 1);
            CHECK_INTEQUAL(bytes_after, 0);
            CHECK_INTEQUAL(*(long *)(void *)prop,
                           RootWindow(display, linux_x11_screen()));
            XFree(prop);
        }
    }

    return 1;
}

/*************************************************************************/
/******************** Test routines: Xinerama support ********************/
/*************************************************************************/

TEST(test_xinerama_disabled_NOINIT)
{
    disable_XineramaQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XineramaQueryExtension = 0;

    Display *display = linux_x11_display();
    CHECK_INTEQUAL(graphics_num_devices(), ScreenCount(display));
    CHECK_INTEQUAL(called_XineramaIsActive, 0);
    CHECK_INTEQUAL(called_XineramaQueryScreens, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xinerama_not_active_NOINIT)
{
    disable_XineramaIsActive = 1;
    CHECK_TRUE(graphics_init());
    disable_XineramaIsActive = 0;

    Display *display = linux_x11_display();
    CHECK_INTEQUAL(graphics_num_devices(), ScreenCount(display));
    CHECK_INTEQUAL(called_XineramaIsActive, 1);
    CHECK_INTEQUAL(called_XineramaQueryScreens, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xinerama_disabled_vidmode_fallback_NOINIT)
{
    disable_XineramaQueryExtension = 1;
    disable_XRRQueryExtension = 1;
    disable_XF86VidModeGetModeLine = 1;
    CHECK_TRUE(graphics_init());
    disable_XineramaQueryExtension = 0;
    disable_XRRQueryExtension = 0;
    disable_XF86VidModeGetModeLine = 0;

    if (!check_vidmode()) {
        SKIP("XF86VidMode not found.");
    }

    Display *display = linux_x11_display();
    CHECK_INTEQUAL(graphics_num_devices(), ScreenCount(display));
    const GraphicsDisplayModeList *mode_list;
    CHECK_TRUE(mode_list = graphics_list_display_modes(1));
    CHECK_INTEQUAL(mode_list->num_modes, graphics_num_devices());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_xinerama_fullscreen_XMoveWindow_NOINIT)
{
    disable_XRRQueryExtension = 1;
    CHECK_TRUE(graphics_init());
    disable_XRRQueryExtension = 0;

    if (!check_xinerama()) {
        SKIP("Xinerama not found.");
    }

    ASSERT(setenv("SIL_X11_FULLSCREEN_METHOD", "XMOVEWINDOW", 1) == 0);
    ASSERT(setenv("SIL_X11_CREATE_FULLSCREEN", "1", 1) == 0);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));

    called_XCreateWindow = 0;
    called_XMoveWindow = 0;
    called_XineramaQueryScreens = 0;
    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(graphics_device_width(),
                                         graphics_device_height(), NULL));
    CHECK_INTEQUAL(called_XCreateWindow, 0);
    CHECK_INTEQUAL(called_XMoveWindow, 1);
    CHECK_INTEQUAL(called_XineramaQueryScreens, 1);

    return 1;
}

/*************************************************************************/
/*********************** Test routines: X11 events ***********************/
/*************************************************************************/

TEST(test_focus_events)
{
    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    Display *display = linux_x11_display();
    Window window = linux_x11_window();

    XSetInputFocus(display, window, RevertToNone, CurrentTime);
    XSync(display, False);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    CHECK_TRUE(graphics_has_focus());

    XSetInputFocus(display, None, RevertToNone, CurrentTime);
    XSync(display, False);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    CHECK_FALSE(graphics_has_focus());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_iconify)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        SKIP("No alternate video mode available.");
    }

    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), mode_width);
    CHECK_INTEQUAL(graphics_display_height(), mode_height);

    int xrandr_width, xrandr_height;
    xrandr_get_current_resolution(&xrandr_width, &xrandr_height);
    CHECK_INTEQUAL(xrandr_width, mode_width);
    CHECK_INTEQUAL(xrandr_height, mode_height);

    /* Iconifying the window should reset to the default screen mode. */
    Display *display = linux_x11_display();
    Window window = linux_x11_window();
    int screen = linux_x11_screen();
    if (!XIconifyWindow(display, window, screen)) {
        SKIP("Window manager does not support iconification.");
    }
    XSync(display, False);
    /* Give the window manager a chance to process the message. */
    nanosleep(&(struct timespec){0, 100*1000*1000}, NULL);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    xrandr_get_current_resolution(&xrandr_width, &xrandr_height);
    CHECK_INTEQUAL(xrandr_width, graphics_device_width());
    CHECK_INTEQUAL(xrandr_height, graphics_device_height());

    /* Restoring the window should restore the custom screen mode. */
    XMapRaised(display, window);
    XSync(display, False);
    nanosleep(&(struct timespec){0, 100*1000*1000}, NULL);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    xrandr_get_current_resolution(&xrandr_width, &xrandr_height);
    CHECK_INTEQUAL(xrandr_width, mode_width);
    CHECK_INTEQUAL(xrandr_height, mode_height);

    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_fullscreen_iconify_set_mode_failure)
{
    if (!check_xrandr()) {
        SKIP("XRandR not found or too old.");
    }

    int mode_width, mode_height;
    if (!get_alternate_mode(&mode_width, &mode_height)) {
        SKIP("No alternate video mode available.");
    }

    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(mode_width, mode_height, NULL));
    CHECK_FALSE(graphics_display_is_window());
    CHECK_INTEQUAL(graphics_display_width(), mode_width);
    CHECK_INTEQUAL(graphics_display_height(), mode_height);

    int xrandr_width, xrandr_height;
    xrandr_get_current_resolution(&xrandr_width, &xrandr_height);
    CHECK_INTEQUAL(xrandr_width, mode_width);
    CHECK_INTEQUAL(xrandr_height, mode_height);

    /* Iconifying the window should reset to the default screen mode. */
    Display *display = linux_x11_display();
    Window window = linux_x11_window();
    int screen = linux_x11_screen();
    if (!XIconifyWindow(display, window, screen)) {
        SKIP("Window manager does not support iconification.");
    }
    XSync(display, False);
    /* Give the window manager a chance to process the message. */
    nanosleep(&(struct timespec){0, 100*1000*1000}, NULL);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    xrandr_get_current_resolution(&xrandr_width, &xrandr_height);
    CHECK_INTEQUAL(xrandr_width, graphics_device_width());
    CHECK_INTEQUAL(xrandr_height, graphics_device_height());

    /* Force a failure of set_video_mode() when the window is restored. */
    disable_XRRGetCrtcInfo = 1;
    XMapRaised(display, window);
    XSync(display, False);
    nanosleep(&(struct timespec){0, 100*1000*1000}, NULL);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    disable_XRRGetCrtcInfo = 0;
    xrandr_get_current_resolution(&xrandr_width, &xrandr_height);
    CHECK_INTEQUAL(xrandr_width, graphics_device_width());
    CHECK_INTEQUAL(xrandr_height, graphics_device_height());
    /* The window should have been changed to non-fullscreen. */
    CHECK_TRUE(graphics_display_is_window());

    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_wm_ping)
{
    Display *display = linux_x11_display();
    Atom wm_protocols = XInternAtom(display, "WM_PROTOCOLS", True);
    Atom net_wm_ping = XInternAtom(display, "_NET_WM_PING", True);
    if (!wm_protocols || !net_wm_ping) {
        SKIP("WM_PROTOCOLS or _NET_WM_PING atom not found.");
    }

    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    Window window = linux_x11_window();
    Window root = RootWindow(display, linux_x11_screen());
    const long timestamp = 12345;
    XClientMessageEvent event = {
        .type = ClientMessage,
        .display = display,
        .window = window,
        .message_type = wm_protocols,
        .format = 32,
        .data = {.l = {net_wm_ping, timestamp, window, 0, 0}},
    };
    CHECK_TRUE(XSendEvent(display, window, False, 0, (XEvent *)&event));
    XSync(display, False);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    CHECK_PTREQUAL(last_event_display, display);
    CHECK_INTEQUAL(last_event_window, root);
    CHECK_FALSE(last_event_propagate);
    CHECK_INTEQUAL(last_event_mask,
                   SubstructureNotifyMask | SubstructureRedirectMask);
    CHECK_INTEQUAL(last_event.type, ClientMessage);
    CHECK_PTREQUAL(last_event.xclient.display, display);
    CHECK_INTEQUAL(last_event.xclient.window, root);
    CHECK_INTEQUAL(last_event.xclient.message_type, wm_protocols);
    CHECK_INTEQUAL(last_event.xclient.format, 32);
    CHECK_INTEQUAL(last_event.xclient.data.l[0], net_wm_ping);
    CHECK_INTEQUAL(last_event.xclient.data.l[1], timestamp);
    CHECK_INTEQUAL(last_event.xclient.data.l[2], window);
    CHECK_INTEQUAL(last_event.xclient.data.l[3], 0);
    CHECK_INTEQUAL(last_event.xclient.data.l[4], 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_unknown_client_message)
{
    Display *display = linux_x11_display();
    Atom wm_protocols = XInternAtom(display, "WM_PROTOCOLS", True);
    Atom net_wm_ping = XInternAtom(display, "_NET_WM_PING", True);

    CHECK_TRUE(graphics_set_display_mode(TESTW, TESTH, NULL));
    Window window = linux_x11_window();
    XClientMessageEvent event = {
        .type = ClientMessage,
        .display = display,
        .window = window,
        .message_type = wm_protocols,
        .format = 32,
        .data = {.l = {0}},
    };

    CHECK_TRUE(XSendEvent(display, window, False, 0, (XEvent *)&event));
    last_event_display = NULL;
    XSync(display, False);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    CHECK_FALSE(last_event_display);  // No event should have been generated.

    event.format = 8;
    event.data.l[0] = net_wm_ping;
    CHECK_TRUE(XSendEvent(display, window, False, 0, (XEvent *)&event));
    last_event_display = NULL;
    XSync(display, False);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    CHECK_FALSE(last_event_display);

    event.format = 32;
    event.message_type = 0;
    CHECK_TRUE(XSendEvent(display, window, False, 0, (XEvent *)&event));
    last_event_display = NULL;
    XSync(display, False);
    while (XPending(display)) {
        XEvent unused_event;
        (void) linux_get_window_event(&unused_event);
    }
    CHECK_FALSE(last_event_display);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // !__clang__

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
